import numpy as np
from numpy.testing import (assert_array_equal, assert_array_almost_equal_nulp,
                           assert_equal, assert_allclose)

from sklearn.datasets import make_regression
from sklearn.linear_model import ElasticNet
from sklearn.metrics import r2_score

from pyuoi import UoI_ElasticNet
from pyuoi.datasets import make_linear_regression


def test_variable_selection():
    """Test basic functionality of UoI_ElasticNet and that it
    finds right model"""

    X, y, w = make_regression(coef=True, random_state=1)
    enet = UoI_ElasticNet(alphas=[1., .9])
    enet.fit(X, y)
    true_coef = np.nonzero(w)[0]
    fit_coef = np.nonzero(enet.coef_)[0]
    assert_array_equal(true_coef, fit_coef)
    assert_array_almost_equal_nulp(true_coef, fit_coef)


def test_estimation_score_usage():
    """Test the ability to change the estimation score in UoI ElasticNet."""

    methods = ('r2', 'AIC', 'AICc', 'BIC')
    X, y = make_regression(n_features=10, n_informative=3,
                           random_state=10)
    scores = []
    for method in methods:
        enet = UoI_ElasticNet(estimation_score=method)
        assert_equal(enet.estimation_score, method)
        enet.fit(X, y)
        y_hat = enet.predict(X)
        assert_equal(r2_score(y, y_hat), enet.score(X, y))
        score = np.max(enet.scores_)
        scores.append(score)
    assert_equal(len(np.unique(scores)), len(methods))


def test_set_random_state():
    """Tests whether random states are handled correctly."""
    X, y = make_regression(n_features=5, n_informative=3,
                           random_state=16, noise=.5)
    # same state
    l1log_0 = UoI_ElasticNet(random_state=13)
    l1log_1 = UoI_ElasticNet(random_state=13)
    l1log_0.fit(X, y)
    l1log_1.fit(X, y)
    assert_array_equal(l1log_0.coef_, l1log_1.coef_)

    # different state
    l1log_1 = UoI_ElasticNet(random_state=14)
    l1log_1.fit(X, y)
    assert not np.array_equal(l1log_0.coef_, l1log_1.coef_)

    # different state, not set
    l1log_0 = UoI_ElasticNet()
    l1log_1 = UoI_ElasticNet()
    l1log_0.fit(X, y)
    l1log_1.fit(X, y)
    assert not np.array_equal(l1log_0.coef_, l1log_1.coef_)


def test_uoi_enet_toy():
    """Test UoI ElasticNet on a toy example."""

    X = np.array([
        [-1, 2],
        [4, 1],
        [1, 3],
        [4, 3],
        [8, 11]], dtype=float)
    beta = np.array([1, 4], dtype=float)
    y = np.dot(X, beta)
    X = np.tile(X, (3, 1))
    y = np.tile(y, 3)

    # choose selection_frac to be slightly smaller to ensure that we get
    # good test sets
    enet = UoI_ElasticNet(
        fit_intercept=False,
        selection_frac=0.75,
        estimation_frac=0.75,
    )
    enet.fit(X, y)

    assert_allclose(enet.coef_, beta)


def test_get_reg_params():
    """Tests whether get_reg_params works correctly for UoI ElasticNet."""

    X = np.array([
        [-1, 2],
        [0, 1],
        [1, 3],
        [4, 3]], dtype=float)
    y = np.array([7, 4, 13, 16], dtype=float)

    # calculate regularization parameters manually
    l1_ratio = .5
    alpha_max = np.max(np.dot(X.T, y) / 4) / l1_ratio
    alphas = [{'alpha': alpha_max, 'l1_ratio': .5},
              {'alpha': alpha_max / 10., 'l1_ratio': .5}]

    # calculate regularization parameters with UoI_ElasticNet object
    enet = UoI_ElasticNet(
        n_lambdas=2,
        fit_intercept=False,
        eps=0.1)
    reg_params = enet.get_reg_params(X, y)

    # check each regularization parameter and key
    for estimate, true in zip(reg_params, alphas):
        assert len(estimate) == len(true)
        for key, value in estimate.items():
            assert_allclose(true[key], value)


def test_intercept_and_coefs_no_selection():
    """Test that UoI Lasso properly calculates the intercept with and without
    standardization."""
    # create line model
    X, y, beta, intercept = make_linear_regression(
        n_samples=500,
        n_features=2,
        n_informative=2,
        snr=10.,
        include_intercept=True,
        random_state=2332)

    # without standardization
    enet = UoI_ElasticNet(
        standardize=False,
        fit_intercept=True
    )
    enet.fit(X, y)
    assert_allclose(enet.intercept_, intercept, rtol=0.25)
    assert_allclose(enet.coef_, beta, rtol=0.25)

    # with standardization
    enet = UoI_ElasticNet(
        standardize=True,
        fit_intercept=True
    )
    enet.fit(X, y)
    assert_allclose(enet.intercept_, intercept, rtol=0.25)
    assert_allclose(enet.coef_, beta, rtol=0.25)


def test_enet_selection_sweep():
    """Tests uoi_selection_sweep for UoI_ElasticNet."""

    # toy data
    X = np.array([
        [-1, 2, 3],
        [4, 1, -7],
        [1, 3, 1],
        [4, 3, 12],
        [8, 11, 2]], dtype=float)
    beta = np.array([1, 4, 2], dtype=float)
    y = np.dot(X, beta)

    # toy regularization
    reg_param_values = [{'alpha': 1.0}, {'alpha': 2.0}]
    enet = UoI_ElasticNet(fit_intercept=True, warm_start=False)
    enet1 = ElasticNet(alpha=1.0, fit_intercept=True, max_iter=enet.max_iter)
    enet2 = ElasticNet(alpha=2.0, fit_intercept=True, max_iter=enet.max_iter)
    enet.output_dim = 1

    coefs = enet.uoi_selection_sweep(X, y, reg_param_values)
    enet1.fit(X, y)
    enet2.fit(X, y)

    assert np.allclose(coefs[0], enet1.coef_)
    assert np.allclose(coefs[1], enet2.coef_)


def test_fit_intercept():
    """Tests whether `include_intercept` in passed through to the linear models.
    """
    enet = UoI_ElasticNet(fit_intercept=True)
    assert enet._selection_lm.fit_intercept
    assert enet._estimation_lm.fit_intercept

    enet = UoI_ElasticNet(fit_intercept=False)
    assert not enet._selection_lm.fit_intercept
    assert not enet._estimation_lm.fit_intercept



import numpy as np


def column_select(V, c, leverage_sort=False, random_state=None):
    """Chooses column indices from a matrix given its SVD.

    Parameters
    ----------
    V : ndarray, shape (n_features, rank)
        The set of singular vectors.

    c : float
        The expected number of columns to select.

    leverage_sort : bool
        If True, resorts the column indices in increasing order of leverage
        score. If False, the column indices are sorted normally.

    Returns
    -------
    column_indices : ndarray of ints
        An array of indices denoting which columns were selected. If
        leverage_sort was true, this array is arranged by increasing leverage
        score.
    """
    # random state
    rng = np.random.RandomState(seed=None)

    # extract number of samples and rank
    n_features, k = V.shape

    # calculate normalized leverage score
    pi = np.sum(V**2, axis=1) / k

    # iterate through columns
    column_flags = np.zeros(n_features, dtype=bool)
    for column in range(n_features):
        # Mahoney (2009), eqn 3
        p = min(1, c * pi[column])
        # selected column randomly
        column_flags[column] = p > rng.rand()

    column_indices = np.argwhere(column_flags).ravel()

    # if desired, sort by increasing leverage score
    if leverage_sort:
        pi_subset = pi[column_indices]
        column_indices = column_indices[np.argsort(pi_subset)]

    return column_indices


def stability_selection_to_threshold(stability_selection, n_boots):
    """Converts user inputted stability selection to an array of
    thresholds. These thresholds correspond to the number of bootstraps
    that a feature must appear in to guarantee placement in the selection
    profile.

    Parameters
    ----------
    stability_selection : int, float, or array-like
        If int, treated as the number of bootstraps that a feature must
        appear in to guarantee placement in selection profile. If float,
        must be between 0 and 1, and is instead the proportion of
        bootstraps.

    n_boots: int
        The number of bootstraps that will be used for selection
    """

    # float, indicating proportion of bootstraps
    if isinstance(stability_selection, float):
        selection_threshold = int(stability_selection * n_boots)

    # int, indicating number of bootstraps
    elif isinstance(stability_selection, int):
        selection_threshold = stability_selection

    else:
        raise ValueError("Stability selection must be a valid float or int.")

    # ensure that ensuing list of selection thresholds satisfies
    # the correct bounds
    if not (
        selection_threshold <= n_boots and selection_threshold >= 1
    ):
        raise ValueError("Stability selection thresholds must be within "
                         "the correct bounds.")

    return selection_threshold


def dissimilarity(H1, H2):
    """Calculates the dissimilarity between two sets of NMF bases.

    Parameters
    ----------
    H1 : ndarray, shape (n_components, n_features)
        First set of bases.

    H2 : ndarray, shape (n_components, n_features)
        Second set of bases.

    Returns
    -------
    diss : float
        Dissimilarity between the two sets of bases.
    """
    k = H1.shape[0]
    H1 = H1 / np.linalg.norm(H1, axis=1, keepdims=True)
    H2 = H2 / np.linalg.norm(H2, axis=1, keepdims=True)
    C = np.dot(H1, H2.T)
    diss = 1 - ((np.max(C, axis=0).sum() + np.max(C, axis=1).sum()) / (2. * k))
    return diss



import numpy as np

from .base import AbstractDecompositionModel

from sklearn.decomposition import TruncatedSVD
from sklearn.model_selection import train_test_split
from sklearn.utils import check_array
from sklearn.utils.validation import check_is_fitted

from .utils import column_select, stability_selection_to_threshold


class UoI_CUR(AbstractDecompositionModel):
    """Performs column subset selection (CUR decomposition) in the Union of
    Intersections framework.

    Parameters
    ----------
    n_boots : int
        Number of bootstraps.
    max_k : int
        The maximum rank of the singular value decomposition.
    boots_frac : float
        The fraction of data to use in the bootstrap.
    algorithm : string, optional
        SVD solver to use. Either “arpack” for the ARPACK wrapper in SciPy
        (scipy.sparse.linalg.svds), or “randomized” for the randomized
        algorithm due to Halko (2009).
    n_iter : int, optional
        Number of iterations for randomized SVD solver. Not used by ARPACK.
        The default is larger than the default in randomized_svd to handle
        sparse matrices that may have large slowly decaying spectrum.
    random_state : int, RandomState instance, or None, optional
        If int, random_state is the seed used by the random number
        generator; If RandomState instance, random_state is the random
        number generator; If None, the random number generator is the
        RandomState instance used by np.random.
    tol : float, optional
        Tolerance for ARPACK. 0 means machine precision. Ignored by
        randomized SVD solver.

    Attributes
    ----------
    components_ : ndarray, shape (n_samples, n_components)
        The selected columns of the design matrix.

    column_indices_ : ndarray, shape (n_components,)
        The indices of the columns selected by the algorithm.
    """
    def __init__(
        self, n_boots, max_k, boots_frac, stability_selection=1.,
        algorithm='randomized', n_iter=5, tol=0.0, random_state=None
    ):
        self.n_boots = n_boots
        self.max_k = max_k
        self.boots_frac = boots_frac
        self.stability_selection = stability_selection
        self.algorithm = algorithm
        self.n_iter = n_iter
        self.tol = tol
        self.random_state = random_state

    def fit(self, X, ks=None, cs=None, stratify=None):
        """Performs column subset selection in the UoI framework on a provided
        matrix.

        Parameters
        ----------
        X : ndarray, shape (n_samples, n_features)
            The data matrix.

        cs : int, float or None
            The expected number of columns to select. If None, c will vary with
            the rank k.

        ks : int, list, ndarray, or None
            The ranks to consider union over. If None, all ranks from
            (1, ..., max_k) will be used.

        stratify : array-like or None
            Ensures groups of samples are alloted to bootstraps proportionally.
            Labels for each group must be an int greater than zero. Must be of
            size equal to the number of samples, with further restrictions on
            the number of groups.

        Returns
        -------
        union : ndarray, shape (n_components,)
            A numpy array containing the indices of the selected columns.
        """
        ks, cs = self.check_ks_and_cs(ks=ks, cs=cs)
        n_ks = ks.size
        n_samples, n_features = X.shape
        selection_threshold = stability_selection_to_threshold(
            self.stability_selection,
            self.n_boots)

        # keep track of which columns are chosen
        columns = np.zeros((n_ks, self.n_boots, n_features), dtype='int')

        # truncated SVD fitter object
        tsvd = TruncatedSVD(n_components=self.max_k,
                            algorithm=self.algorithm,
                            n_iter=self.n_iter,
                            tol=self.tol,
                            random_state=self.random_state)

        # iterate over bootstraps
        for bootstrap in range(self.n_boots):
            # extract bootstrap
            if self.boots_frac == 1:
                X_train = X
            else:
                X_train, _ = train_test_split(X, test_size=1 - self.boots_frac,
                                              stratify=stratify,
                                              random_state=self.random_state)
            # perform truncated SVD on bootstrap
            tsvd.fit(X_train)
            # extract right singular vectors
            V = tsvd.components_.T

            # iterate over ranks
            for k_idx, k in enumerate(ks):
                c = cs[k_idx]

                column_indices = column_select(V[:, :k], c=c,
                                               random_state=self.random_state)
                # convert column flags to set
                columns[k_idx, bootstrap][column_indices] += 1

        intersection = np.sum(columns, axis=1) >= selection_threshold

        # calculate union of intersections
        union = np.sum(intersection, axis=0) > 0
        self.column_indices_ = np.argwhere(union).ravel()

        if self.column_indices_.size == 0:
            self.components_ = np.array([[]])
        else:
            self.components_ = X[:, self.column_indices_]
        return self

    def check_ks_and_cs(self, ks=None, cs=None):
        """Process the set of ranks to calculate leverage scores over, and the
        expected number of columns for each rank.

        Parameters
        ----------
        ks : ndarray
            The ranks to compute leverage scores over.
        cs : ndarray
            The expected number of columns to select for each rank.

        Returns
        -------
        ks : ndarray
            Processed and checked ranks.
        cs : ndarray
            Processed expected number of columns.
        """
        # convert ks to a numpy array
        if ks is None:
            ks = 1 + np.arange(self.max_k)
        elif isinstance(ks, int):
            ks = np.array([ks])
        elif isinstance(ks, list):
            ks = np.array(ks)
        elif not isinstance(ks, np.ndarray):
            raise ValueError('ks must be a valid int, list or numpy array.')

        # check that the numpy array contains valid values
        if (
            (not np.all(ks <= self.max_k))
            or (not np.all(ks > 0))
            or (not np.issubdtype(ks.dtype, np.integer))
        ):
            raise ValueError('Ranks must be positive, integers, and no more'
                             ' than the max rank.')

        # check expected column numbers
        n_ks = ks.size

        if cs is None:
            cs = ks + 20
        elif isinstance(cs, int) or isinstance(cs, float):
            cs = cs * np.ones(n_ks)
        elif isinstance(cs, list):
            cs = np.array(cs)
        elif not isinstance(cs, np.ndarray):
            raise ValueError('cs must be a valid int, float, list or numpy'
                             ' array.')

        # check that the numpy array contains valid values
        if (
            (not np.all(cs > 0))
            or (not np.issubdtype(cs.dtype, np.number))
            or (not cs.size == ks.size)
        ):
            raise ValueError('Expected number of columns must consist of'
                             ' positive numbers, with total size equal to ks.')
        return ks, cs

    def transform(self, X):
        """Transform the data by extracting the selected columns.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Data matrix from which to select columns.

        Returns
        -------
        X_new : array-like, shape (n_samples, n_components)
            Data matrix comprised of selected columns.
        """
        check_is_fitted(self, ['components_'])
        X = check_array(X)
        X_new = X[:, self.column_indices_]
        return X_new

    def fit_transform(self, X, ks=None, cs=None):
        """Fit and transform the data by choosing and extracting specific
        columns.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Data matrix from which to select columns.

        Returns
        -------
        X_new : array-like, shape (n_samples, n_components)
            Data matrix comprised of selected columns.
        """
        self.fit(X, ks=ks, cs=cs)
        return self.transform(X)


class CUR(AbstractDecompositionModel):
    """Performs ordinary column subset selection through a CUR decomposition.

    Parameters
    ----------
    max_k : int
        The maximum rank of the singular value decomposition.
    algorithm : string, optional
        SVD solver to use. Either “arpack” for the ARPACK wrapper in SciPy
        (scipy.sparse.linalg.svds), or “randomized” for the randomized
        algorithm due to Halko (2009).
    n_iter : int, optional
        Number of iterations for randomized SVD solver. Not used by ARPACK.
        The default is larger than the default in randomized_svd to handle
        sparse matrices that may have large slowly decaying spectrum.
    random_state : int, RandomState instance, or None, optional
        If int, random_state is the seed used by the random number
        generator; If RandomState instance, random_state is the random
        number generator; If None, the random number generator is the
        RandomState instance used by np.random.
    tol : float, optional
        Tolerance for ARPACK. 0 means machine precision. Ignored by
        randomized SVD solver.

    Attributes
    ----------
    components_ : ndarray, shape (n_samples, n_components)
        The selected columns of the design matrix.
    column_indices_ : ndarray, shape (n_components,)
        The indices of the columns selected by the algorithm.
    """
    def __init__(
        self, max_k, algorithm='randomized', n_iter=5, tol=0.0,
        random_state=None
    ):
        self.max_k = max_k
        self.algorithm = algorithm
        self.n_iter = n_iter
        self.tol = tol
        self.random_state = random_state

    def fit(self, X, c=None):
        """Performs column subset selection in the UoI framework on a provided
        matrix.

        Parameters
        ----------
        X : ndarray, shape (n_samples, n_features)
            The data matrix.
        c : float
            The expected number of columns to select. If None, c will vary with
            the rank k.
        """
        n_samples, n_features = X.shape

        # truncated SVD fitter object
        tsvd = TruncatedSVD(n_components=self.max_k,
                            algorithm=self.algorithm,
                            n_iter=self.n_iter,
                            tol=self.tol,
                            random_state=self.random_state)

        # perform truncated SVD on bootstrap
        tsvd.fit(X)
        # extract right singular vectors
        V = tsvd.components_.T[:, :self.max_k]

        # perform column selection on the subset of singular vectors
        if c is None:
            c = self.max_k + 20

        column_indices = column_select(V=V, c=c,
                                       leverage_sort=True,
                                       random_state=self.random_state)

        self.column_indices_ = column_indices
        self.components_ = X[:, self.column_indices_]
        return self

    def transform(self, X):
        """Transform the data by extracting the selected columns.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Data matrix from which to select columns.

        Returns
        -------
        X_new : array-like, shape (n_samples, n_components)
            Data matrix comprised of selected columns.
        """
        check_is_fitted(self, ['column_indices_'])
        X = check_array(X)
        X_new = X[:, self.column_indices_]
        return X_new

    def fit_transform(self, X, c=None):
        """Fit and transform the data by choosing and extracting specific
        columns.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Data matrix from which to select columns.

        Returns
        -------
        X_new : array-like, shape (n_samples, n_components)
            Data matrix comprised of selected columns.
        """
        self.fit(X, c=c)
        return self.transform(X)



import numpy as np
from sklearn.utils import check_random_state

from ..utils import softmax, sigmoid


def load_swimmer(flatten=True):
    from pkg_resources import resource_filename
    import h5py
    with h5py.File(resource_filename('pyuoi', 'data/Swimmer.h5'), 'r+') as f:
        swimmers = f['Y'][:].astype(float)
    if flatten:
        swimmers = swimmers.T.reshape(256, 1024)
    return swimmers


def make_linear_regression(n_samples=100, n_features=5, n_informative=2,
                           X_loc=3., X_scale=1., snr=5.,
                           beta=None, beta_low=1., beta_high=3.,
                           include_intercept=False, random_state=None):
    """Make a Linear regression dataset.

    Parameters
    ----------
    n_samples : int
        The number of samples to make.
    n_features : int
        The number of feature to use.
    n_informative : int
        The number of feature with non-zero weights.
    X_loc : float
        The mean of the features in the design matrix.
    X_scale : float
        The standard deviation of the features in the design matrix.
    snr : float
        The signal-to-noise ratio, which informs the variance of the noise
        term.
    beta : np.ndarray or None
        The beta values to use. If None, beta values will be drawn from a
        uniform distribution.
    beta_low : float
        The lower bound for the beta values.
    beta_high : float
        The upper bound for the beta values.
    include_intercept : bool
        If true, includes an intercept in the model, if False, the intercept is
        set to 0.
    random_state : int, np.random.RandomState instance, or None
        Random number seed or state.

    Returns
    -------
    X : ndarray, shape (n_samples, n_features)
        The design matrix.
    y : ndarray, shape (n_samples,)
        The response vector.
    beta : ndarray, shape (n_features,)
        The feature coefficients.
    intercept : float
        The intercept. If include_intercept is False, then intercept is zero.
    """
    rng = check_random_state(random_state)

    # create design matrix
    X = rng.normal(loc=X_loc,
                   scale=X_scale,
                   size=(n_samples, n_features))

    # create coefficients
    if beta is None:
        # draw beta values from gamma distribution
        beta = rng.uniform(low=beta_low,
                           high=beta_high,
                           size=n_features)

        # choose sparsity mask
        zero_idx = np.zeros(n_features)
        zero_idx[:n_informative] = 1
        rng.shuffle(zero_idx)
        # randomly assign beta elements to zero
        beta = beta * zero_idx

    # create intercept
    if include_intercept:
        intercept = rng.uniform(low=beta_low, high=beta_high)
    else:
        intercept = 0

    # draw response variable
    eta = intercept + np.dot(X, beta)
    signal_var = np.var(eta)
    noise_var = signal_var / snr
    noise = rng.normal(loc=0, scale=np.sqrt(noise_var), size=eta.shape)
    y = eta + noise

    return X, y, beta, intercept


def make_classification(n_samples=100, n_features=20, n_informative=2,
                        n_classes=2, shared_support=False, random_state=None,
                        w_scale=1., include_intercept=False):
    """Make a linear classification dataset.

    Parameters
    ----------
    n_samples : int
        The number of samples to make.
    n_features : int
        The number of feature to use.
    n_informative : int
        The number of feature with non-zero weights.
    n_classes : int
        The number of classes.
    shared_support : bool
        If True, all classes will share the same random support. If False, they
        will each have randomly chooses support.
    random_state : int or np.random.RandomState instance
        Random number seed or state.
    w_scale : float
        The model parameter matrix, w, will be drawn from a normal distribution
        with std=w_scale.
    include_intercept : bool
        If true, includes an intercept in the model, if False, the intercept is
        set to 0.
    """
    if isinstance(random_state, int):
        rng = np.random.RandomState(random_state)
    else:
        if random_state is None:
            rng = np.random
        else:
            rng = random_state
    n_not_informative = n_features - n_informative

    X = rng.randn(n_samples, n_features)
    X -= X.mean(axis=-1, keepdims=True)
    X /= X.std(axis=-1, keepdims=True)

    if n_classes > 2:
        w = rng.randn(n_features, n_classes)
        if include_intercept:
            intercept = rng.randn(1, n_classes)
            intercept -= intercept.max()
        else:
            intercept = np.zeros((1, n_classes))
        if n_not_informative > 0:
            if shared_support:
                idxs = rng.permutation(n_features)[:n_not_informative]
                w[idxs] = 0.
            else:
                for ii in range(n_classes):
                    idxs = rng.permutation(n_features)[:n_not_informative]
                    w[idxs, ii * np.ones_like(idxs, dtype=int)] = 0.
    else:
        w = rng.randn(n_features, 1)
        if include_intercept:
            intercept = rng.randn(1, 1)
        else:
            intercept = np.zeros((1, 1))
        if n_not_informative > 0:
            idxs = rng.permutation(n_features)[:n_not_informative]
            w[idxs] = 0.
    w *= w_scale

    log_p = X.dot(w)
    if include_intercept:
        log_p += intercept
    if n_classes > 2:
        p = softmax(log_p)
        y = np.array([rng.multinomial(1, pi) for pi in p])
        y = y.argmax(axis=-1)
    else:
        p = sigmoid(np.squeeze(log_p))
        y = np.array([rng.binomial(1, pi) for pi in p])

    return X, y, w.T, intercept


def make_poisson_regression(n_samples=100, n_features=5, n_informative=2,
                            X_loc=0., X_scale=1. / 8,
                            beta=None, beta_shape=1., beta_scale=3.,
                            include_intercept=False, random_state=None):
    """Make a Poisson regression dataset.

    Parameters
    ----------
    n_samples : int
        The number of samples to make.
    n_features : int
        The number of feature to use.
    n_informative : int
        The number of feature with non-zero weights.
    X_loc : float
        The mean of the features in the design matrix.
    X_scale : float
        The standard deviation of the features in the design matrix.
    beta : np.ndarray or None
        The beta values to use. If None, beta values will be drawn from a gamma
        distribution.
    beta_shape : float
        The shape parameter for the beta values.
    beta_scale : float
        The scale parameter for the beta values.
    include_intercept : bool
        If true, includes an intercept in the model, if False, the intercept is
        set to 0.
    random_state : int, np.random.RandomState instance, or None
        Random number seed or state.

    Returns
    -------
    X : ndarray, shape (n_samples, n_features)
        The design matrix.
    y : ndarray, shape (n_samples,)
        The response vector.
    beta : ndarray, shape (n_features,)
        The feature coefficients.
    intercept : float
        The intercept. If include_intercept is False, then intercept is zero.
    """
    rng = check_random_state(random_state)

    # create design matrix
    X = rng.normal(loc=X_loc,
                   scale=X_scale,
                   size=(n_samples, n_features))

    # create coefficients
    if beta is None:
        # draw beta values from gamma distribution
        beta = rng.gamma(shape=beta_shape,
                         scale=beta_scale,
                         size=n_features)
        # choose sparsity mask
        zero_idx = np.zeros(n_features)
        zero_idx[:n_informative] = 1
        rng.shuffle(zero_idx)
        # randomly assign beta elements to zero
        beta = beta * zero_idx

    # create intercept
    if include_intercept:
        intercept = rng.gamma(shape=beta_shape, scale=beta_scale)
    else:
        intercept = 0

    # draw response variable
    eta = intercept + np.dot(X, beta)
    y = rng.poisson(np.exp(eta))

    return X, y, beta, intercept



import pytest
import numpy as np

from numpy.testing import assert_array_equal, assert_raises
from pyuoi.decomposition import UoI_NMF, UoI_NMF_Base
from pyuoi.decomposition.utils import dissimilarity
from sklearn.cluster import DBSCAN
from sklearn.decomposition import NMF


@pytest.fixture
def nmf_setup():
    W = np.random.randint(0, high=2, size=(500, 5))
    H = np.random.randint(0, high=2, size=(5, 2))
    X = np.dot(W, H)
    noise = np.random.normal(loc=0, scale=0.5, size=X.shape)**2
    X = X + noise
    return X


def test_dissimilarity():
    """Test the dissimilarity function."""
    k = 5
    n_features = 20

    # same bases should be a dissimilarity of zero
    H1 = np.random.randint(low=0, high=3, size=(k, n_features))
    H2 = np.copy(H1)
    assert np.allclose(dissimilarity(H1, H2), 0.)


@pytest.mark.fast
def test_UoI_NMF_Base_initialization():
    """Tests the initialization of UoI NMF Base Class."""
    n_boots = 30
    ranks = 10
    uoi = UoI_NMF(n_boots=n_boots, ranks=ranks,
                  random_state=np.random.RandomState(2332))
    assert_array_equal(uoi.ranks, np.arange(2, ranks + 1))
    assert uoi.nmf.solver == 'mu'
    assert uoi.nmf.beta_loss == 'kullback-leibler'
    assert uoi.cluster.min_samples == n_boots / 2


@pytest.mark.fast
def test_UoI_NMF_initialization():
    """Tests the initialization of UoI NMF."""
    n_boots = 30
    ranks = 10
    uoi = UoI_NMF(n_boots=n_boots, ranks=ranks)
    assert_array_equal(uoi.ranks, np.arange(2, ranks + 1))
    assert uoi.nmf.solver == 'mu'
    assert uoi.nmf.beta_loss == 'kullback-leibler'
    assert uoi.cluster.min_samples == n_boots / 2
    assert uoi.cons_meth == np.mean


@pytest.mark.fast
def test_UoI_NMF_initialization_value_error():
    """Tests that ValueErrors are correctly raised in the NMF initialization."""
    assert_raises(ValueError, UoI_NMF_Base, **{'ranks': 2.5})
    assert_raises(ValueError, UoI_NMF_Base, **{'nmf': NMF})
    assert_raises(ValueError, UoI_NMF_Base, **{'cluster': DBSCAN})
    assert_raises(ValueError, UoI_NMF_Base, **{'nnreg': 2})


@pytest.mark.fast
def test_UoI_NMF_fit(nmf_setup):
    """Tests that the fitting procedure of UoI NMF runs without error."""
    X = nmf_setup

    n_boots = 1
    ranks = 5
    uoi = UoI_NMF(n_boots=n_boots,
                  ranks=[ranks],
                  nmf_max_iter=1000,
                  random_state=2332,
                  use_dissimilarity=False)
    uoi.fit(X)
    assert hasattr(uoi, 'components_')


@pytest.mark.fast
def test_UoI_NMF_fit_no_dissimilarity(nmf_setup):
    """Tests that the fitting procedure of UoI NMF runs without error, when
    the algorithm does not use dissimilarity to choose a rank."""
    X = nmf_setup

    n_boots = 1
    ranks = 5
    uoi = UoI_NMF(n_boots=n_boots,
                  ranks=[ranks],
                  nmf_max_iter=1000,
                  random_state=2332,
                  use_dissimilarity=False)
    uoi.fit(X)
    assert hasattr(uoi, 'components_')


@pytest.mark.fast
def test_UoI_NMF_transform(nmf_setup):
    """Tests that the transform procedure of UoI NMF runs without error."""
    X = nmf_setup

    n_boots = 1
    ranks = 5
    uoi = UoI_NMF(n_boots=n_boots,
                  ranks=[ranks],
                  nmf_max_iter=1000,
                  random_state=2332,
                  use_dissimilarity=False)
    X_tfm = uoi.fit_transform(X)
    assert hasattr(uoi, 'components_')
    assert X_tfm is not None


@pytest.mark.fast
def test_UoI_NMF_transform_value_error(nmf_setup):
    """Tests that the transform procedure of UoI NMF correctly raises a
    ValueError."""
    X = nmf_setup
    n_boots = 1
    ranks = 5
    uoi = UoI_NMF(n_boots=n_boots,
                  ranks=[ranks],
                  nmf_max_iter=1000,
                  random_state=2332,
                  use_dissimilarity=False)
    uoi.fit(X)

    # transform
    Y = np.random.normal(size=(X.shape[0], 2 * X.shape[1]))**2
    assert_raises(ValueError, uoi.transform, Y)
    # inverse transform
    W = np.random.normal(size=(X.shape[0], 2 * uoi.components_.shape[0]))**2
    assert_raises(ValueError, uoi.inverse_transform, W)


@pytest.mark.fast
def test_UoI_NMF_reconstruction_error(nmf_setup):
    """Tests that a reconstruction error is calculated when data is
    transformed."""
    X = nmf_setup
    n_boots = 1
    ranks = 5
    uoi = UoI_NMF(n_boots=n_boots,
                  ranks=[ranks],
                  nmf_max_iter=1000,
                  random_state=2332,
                  use_dissimilarity=False)
    uoi.fit(X)
    X_tfm = uoi.transform(X, reconstruction_err=True)
    assert hasattr(uoi, 'components_')
    assert hasattr(uoi, 'reconstruction_err_')
    assert uoi.reconstruction_err_ is not None
    assert X_tfm is not None


@pytest.mark.slow
def test_UoI_NMF_correct_number_of_components():
    """Tests that, using the dissimilarity metric, UoI NMF extracts the correct
    number of bases."""
    k = 2
    n_samples = 1000
    n_features = 30

    # create data matrix
    W = np.random.randint(low=0, high=3, size=(n_samples, k))
    H = np.random.randint(low=0, high=3, size=(k, n_features))
    noise = np.random.normal(loc=0, scale=0.5, size=(n_samples, n_features))**2
    A = np.dot(W, H) + noise

    # fit uoi nmf
    uoi = UoI_NMF(n_boots=10,
                  ranks=[2, 4, 8],
                  nmf_max_iter=5000,
                  use_dissimilarity=True)
    uoi.fit(A)

    assert uoi.components_.shape[0] == k


@pytest.mark.fast
def test_UoI_NMF_dissim_boots_argcheck(nmf_setup):
    """Test that UoI_NMF raises ValueError when trying to use
    dissimilarity with a single bootstrap."""
    n_boots = 1
    ranks = 5
    assert_raises(ValueError, UoI_NMF,
                  n_boots=n_boots,
                  ranks=[ranks],
                  nmf_max_iter=1000,
                  random_state=2332)



import pytest, numbers, warnings
import numpy as np

from numpy.testing import assert_array_equal, assert_allclose, assert_equal

from scipy.sparse import rand as sprand
from scipy import optimize

from pyuoi import UoI_L1Logistic
from pyuoi.linear_model.logistic import (fit_intercept_fixed_coef,
                                         MaskedCoefLogisticRegression,
                                         LogisticInterceptFitterNoFeatures,
                                         _logistic_regression_path,
                                         _multinomial_loss_grad,
                                         _logistic_loss_and_grad)
from sklearn.metrics import accuracy_score
from sklearn.preprocessing import LabelEncoder, OneHotEncoder
from sklearn.utils import (compute_class_weight,
                           check_consistent_length, check_array)
from sklearn.exceptions import ConvergenceWarning

from pyuoi.datasets import make_classification
from pyuoi.lbfgs import fmin_lbfgs, AllZeroLBFGSError


def _logistic_regression_path_old(X, y, Cs=48, fit_intercept=True,
                                  max_iter=100, tol=1e-4, verbose=0, coef=None,
                                  class_weight=None, penalty='l2',
                                  multi_class='auto',
                                  check_input=True,
                                  sample_weight=None,
                                  l1_ratio=None, coef_mask=None):
    """Compute a Logistic Regression model for a list of regularization
    parameters.

    This is the original function used to check the new indexing-based
    version rather than the masking version implemented here.

    Parameters
    ----------
    X : array-like or sparse matrix, shape (n_samples, n_features)
        Input data.
    y : array-like, shape (n_samples,) or (n_samples, n_targets)
        Input data, target values.
    Cs : int | array-like, shape (n_cs,)
        List of values for the regularization parameter or integer specifying
        the number of regularization parameters that should be used. In this
        case, the parameters will be chosen in a logarithmic scale between
        1e-4 and 1e4.
    fit_intercept : bool
        Whether to fit an intercept for the model. In this case the shape of
        the returned array is (n_cs, n_features + 1).
    max_iter : int
        Maximum number of iterations for the solver.
    tol : float
        Stopping criterion. For the newton-cg and lbfgs solvers, the iteration
        will stop when ``max{|g_i | i = 1, ..., n} <= tol``
        where ``g_i`` is the i-th component of the gradient.
    verbose : int
        For the liblinear and lbfgs solvers set verbose to any positive
        number for verbosity.
    coef : array-like, shape (n_features,), default None
        Initialization value for coefficients of logistic regression.
        Useless for liblinear solver.
    class_weight : dict or 'balanced', optional
        Weights associated with classes in the form ``{class_label: weight}``.
        If not given, all classes are supposed to have weight one.
        The "balanced" mode uses the values of y to automatically adjust
        weights inversely proportional to class frequencies in the input data
        as ``n_samples / (n_classes * np.bincount(y))``.
        Note that these weights will be multiplied with sample_weight (passed
        through the fit method) if sample_weight is specified.
    multi_class : str, {'multinomial', 'auto'}, default: 'auto'
        For 'multinomial' the loss minimised is the multinomial loss fit
        across the entire probability distribution, *even when the data is
        binary*. 'auto' selects binary if the data is binary
        and otherwise selects 'multinomial'.
    check_input : bool, default True
        If False, the input arrays X and y will not be checked.
    sample_weight : array-like, shape(n_samples,) optional
        Array of weights that are assigned to individual samples.
        If not provided, then each sample is given unit weight.
    coef_mask : array-like, shape (n_features), (n_classes, n_features) optional
        Masking array for coef.
    Returns
    -------
    coefs : ndarray, shape (n_cs, n_features) or (n_cs, n_features + 1)
        List of coefficients for the Logistic Regression model. If
        fit_intercept is set to True then the second dimension will be
        n_features + 1, where the last item represents the intercept. For
        ``multiclass='multinomial'``, the shape is (n_classes, n_cs,
        n_features) or (n_classes, n_cs, n_features + 1).
    Cs : ndarray
        Grid of Cs used for cross-validation.
    n_iter : array, shape (n_cs,)
        Actual number of iteration for each Cs.
    """
    if isinstance(Cs, numbers.Integral):
        Cs = np.logspace(-4, 4, Cs)

    # Preprocessing.
    if check_input:
        X = check_array(X, accept_sparse='csr', dtype=np.float64,
                        accept_large_sparse=True)
        y = check_array(y, ensure_2d=False, dtype=None)
        check_consistent_length(X, y)
    _, n_features = X.shape

    classes = np.unique(y)

    if multi_class == 'auto':
        if len(classes) > 2:
            multi_class = 'multinomial'
        else:
            multi_class = 'ovr'

    # If sample weights exist, convert them to array (support for lists)
    # and check length
    # Otherwise set them to 1 for all examples
    if sample_weight is not None:
        sample_weight = np.array(sample_weight, dtype=X.dtype, order='C')
        check_consistent_length(y, sample_weight)
    else:
        sample_weight = np.ones(X.shape[0], dtype=X.dtype)

    # If class_weights is a dict (provided by the user), the weights
    # are assigned to the original labels. If it is "balanced", then
    # the class_weights are assigned after masking the labels with a OvR.
    le = LabelEncoder()
    if isinstance(class_weight, dict) or multi_class == 'multinomial':
        class_weight_ = compute_class_weight(class_weight, classes=classes, y=y)
        sample_weight *= class_weight_[le.fit_transform(y)]

    # For doing a ovr, we need to mask the labels first. for the
    # multinomial case this is not necessary.
    if multi_class == 'ovr':
        coef_size = n_features
        w0 = np.zeros(n_features + int(fit_intercept), dtype=X.dtype)
        mask_classes = np.array([-1, 1])
        mask = (y == 1)
        y_bin = np.ones(y.shape, dtype=X.dtype)
        y_bin[~mask] = -1.
        # for compute_class_weight

        if class_weight == "balanced":
            class_weight_ = compute_class_weight(class_weight,
                                                 classes=mask_classes,
                                                 y=y_bin)
            sample_weight *= class_weight_[le.fit_transform(y_bin)]

    else:
        coef_size = classes.size * n_features
        lbin = OneHotEncoder(categories=[range(classes.size)], sparse=False)
        Y_multi = lbin.fit_transform(y[:, np.newaxis])
        if Y_multi.shape[1] == 1:
            Y_multi = np.hstack([1 - Y_multi, Y_multi])
        w0 = np.zeros((classes.size, n_features + int(fit_intercept)),
                      dtype=X.dtype)
        w0[:, -1] = LogisticInterceptFitterNoFeatures(y,
                                                      classes.size).intercept_

    if coef is not None:
        # it must work both giving the bias term and not
        if multi_class == 'ovr':
            if coef.size not in (n_features, w0.size):
                raise ValueError(
                    'Initialization coef is of shape %d, expected shape '
                    '%d or %d' % (coef.size, n_features, w0.size))
            w0[:coef.size] = coef
        else:
            w0[:, :coef.shape[1]] = coef

    # Mask initial array
    if coef_mask is not None:
        if multi_class == 'ovr':
            w0[:n_features] *= coef_mask
        else:
            w0[:, :n_features] *= coef_mask

    if multi_class == 'multinomial':
        # fmin_l_bfgs_b and newton-cg accepts only ravelled parameters.
        target = Y_multi
        if penalty == 'l2':
            w0 = w0.ravel()

            def func(x, *args):
                return _multinomial_loss_grad(x, *args)[0:2]
        else:
            w0 = w0.T.ravel().copy()

            def inner_func(x, *args):
                return _multinomial_loss_grad(x, *args)[0:2]

            def func(x, g, *args):
                x = x.reshape(-1, classes.size).T.ravel()
                loss, grad = inner_func(x, *args)
                grad = grad.reshape(classes.size, -1).T.ravel()
                g[:] = grad
                return loss
    else:
        target = y_bin
        if penalty == 'l2':
            func = _logistic_loss_and_grad
        else:
            def func(x, g, *args):
                loss, grad = _logistic_loss_and_grad(x, *args)
                g[:] = grad
                return loss

    coefs = list()
    n_iter = np.zeros(len(Cs), dtype=np.int32)
    for i, C in enumerate(Cs):
        iprint = [-1, 50, 1, 100, 101][
            np.searchsorted(np.array([0, 1, 2, 3]), verbose)]
        if penalty == 'l2':
            w0, loss, info = optimize.fmin_l_bfgs_b(
                func, w0, fprime=None,
                args=(X, target, 1. / C, coef_mask, sample_weight),
                iprint=iprint, pgtol=tol, maxiter=max_iter)
        else:
            zeros_seen = [0]

            def zero_coef(x, *args):
                if multi_class == 'multinomial':
                    x = x.reshape(-1, classes.size)[:-1]
                else:
                    x = x[:-1]
                now_zeros = np.array_equiv(x, 0.)
                if now_zeros:
                    zeros_seen[0] += 1
                else:
                    zeros_seen[0] = 0
                if zeros_seen[0] > 1:
                    return -2048
            try:
                w0 = fmin_lbfgs(func, w0, orthantwise_c=1. / C,
                                args=(X, target, 0., coef_mask, sample_weight),
                                max_iterations=max_iter,
                                epsilon=tol,
                                orthantwise_end=coef_size,
                                progress=zero_coef)
            except AllZeroLBFGSError:
                w0 *= 0.
            info = None
        if info is not None and info["warnflag"] == 1:
            warnings.warn("lbfgs failed to converge. Increase the number "
                          "of iterations.", ConvergenceWarning)
        # In scipy <= 1.0.0, nit may exceed maxiter.
        # See https://github.com/scipy/scipy/issues/7854.
        if info is None:
            n_iter_i = -1
        else:
            n_iter_i = min(info['nit'], max_iter)

        if multi_class == 'multinomial':
            n_classes = max(2, classes.size)
            if penalty == 'l2':
                multi_w0 = np.reshape(w0, (n_classes, -1))
            else:
                multi_w0 = np.reshape(w0, (-1, n_classes)).T
            if coef_mask is not None:
                multi_w0[:, :n_features] *= coef_mask
            coefs.append(multi_w0.copy())
        else:
            if coef_mask is not None:
                w0[:n_features] *= coef_mask
            coefs.append(w0.copy())

        n_iter[i] = n_iter_i

    return np.array(coefs), np.array(Cs), n_iter


def test_fit_intercept_fixed_coef():
    """Test that the intercept in fit correctly for fixed coefficients."""
    X = np.zeros((6, 5))
    coef = np.ones((1, 5))
    y = np.ones(6, dtype=int)
    y[:3] = 0
    b = fit_intercept_fixed_coef(X, coef, y, 2)
    assert_allclose(b, 0.)

    X = np.zeros((7, 5))
    y = np.ones(7, dtype=int)
    y[:3] = 0
    b = fit_intercept_fixed_coef(X, coef, y, 3)
    assert_allclose(b.argmax(), 1)
    assert_allclose(b.argmin(), 2)


def test_fit_intercept_no_features():
    """Test that the intercept in fit correctly for fixed coefficients."""
    X = np.zeros((5, 1))
    y = np.ones(6, dtype=int)
    y[:3] = 0
    LR = LogisticInterceptFitterNoFeatures(y, 1)
    b = LR.intercept_
    assert_allclose(b, 0.)

    y = np.ones(7, dtype=int)
    y[:3] = 0
    LR = LogisticInterceptFitterNoFeatures(y, 1)
    yhat = LR.predict(X)
    assert_allclose(yhat, 1)
    py = LR.predict_proba(X)
    assert np.all(py > .5)

    y = np.ones(7, dtype=int)
    y[:3] = 0
    LR = LogisticInterceptFitterNoFeatures(y, 3)

    yhat = LR.predict(X)
    assert_allclose(yhat, 1)

    py = LR.predict_proba(X)
    assert_allclose(py.argmax(axis=1), 1)
    assert_allclose(py.argmin(axis=1), 2)


def test_l1logistic_intercept():
    """Test that binary L1 Logistic fits an intercept when run."""
    for fi in [True, False]:
        X, y, w, b = make_classification(n_samples=100,
                                         random_state=11,
                                         n_features=4,
                                         w_scale=4.,
                                         include_intercept=fi)
        l1log = UoI_L1Logistic(fit_intercept=fi,
                               n_boots_sel=3,
                               n_boots_est=3).fit(X, y)
        if not fi:
            assert_array_equal(l1log.intercept_, 0.)
        else:
            l1log.intercept_


def test_l1logistic_binary():
    """Test that binary L1 Logistic runs in the UoI framework."""
    n_inf = 10
    X, y, w, b = make_classification(n_samples=200,
                                     random_state=6,
                                     n_informative=n_inf,
                                     n_features=20,
                                     w_scale=4.,
                                     include_intercept=True)

    l1log = UoI_L1Logistic(random_state=10).fit(X, y)
    l1log = UoI_L1Logistic(random_state=10, fit_intercept=False).fit(X, y)
    l1log.predict_proba(X)
    l1log.predict_log_proba(X)
    y_hat = l1log.predict(X)
    assert_equal(accuracy_score(y, y_hat), l1log.score(X, y))
    assert (np.sign(abs(w)) == np.sign(abs(l1log.coef_))).mean() >= .8


def test_l1logistic_binary_multinomial():
    """Test that binary L1 Logistic runs in the UoI framework
    using multi_class='multinomial'."""
    n_inf = 10
    X, y, w, b = make_classification(n_samples=200,
                                     random_state=6,
                                     n_informative=n_inf,
                                     n_features=20,
                                     w_scale=4.,
                                     include_intercept=True)

    UoI_L1Logistic(random_state=10, multi_class='multinomial').fit(X, y)
    UoI_L1Logistic(random_state=10, fit_intercept=False,
                   multi_class='multinomial').fit(X, y)


def test_l1logistic_no_ovr():
    """Test that binary L1 Logistic model raises an error for
    multiclass='ovr'."""
    with pytest.raises(ValueError):
        UoI_L1Logistic(multi_class='ovr')


def test_l1logistic_multiclass():
    """Test that multiclass L1 Logistic runs in the UoI framework when all
       classes share a support."""
    n_features = 20
    n_inf = 10
    X, y, w, b = make_classification(n_samples=200,
                                     random_state=10,
                                     n_classes=5,
                                     n_informative=n_inf,
                                     n_features=n_features,
                                     shared_support=True,
                                     w_scale=4.)
    l1log = UoI_L1Logistic().fit(X, y)
    l1log.predict_proba(X)
    l1log.predict_log_proba(X)
    y_hat = l1log.predict(X)
    assert_equal(accuracy_score(y, y_hat), l1log.score(X, y))
    assert (np.sign(abs(w)) == np.sign(abs(l1log.coef_))).mean() >= .8


def test_l1logistic_multiclass_not_shared():
    """Test that multiclass L1 Logistic runs in the UoI framework when all
       classes share a support."""
    n_features = 20
    n_inf = 10
    X, y, w, b = make_classification(n_samples=400,
                                     random_state=10,
                                     n_classes=5,
                                     n_informative=n_inf,
                                     n_features=n_features,
                                     shared_support=False,
                                     w_scale=4.)
    l1log = UoI_L1Logistic(shared_support=False).fit(X, y)
    l1log.predict_log_proba(X)
    y_hat = l1log.predict(X)
    assert_equal(accuracy_score(y, y_hat), l1log.score(X, y))
    assert (np.sign(abs(w)) == np.sign(abs(l1log.coef_))).mean() >= .7


def test_masked_logistic():
    """Test the masked logistic regression class."""
    n_features = 20
    n_inf = 10
    for shared_support in [True, False]:
        for n_classes in [2, 3]:
            for intercept in [True, False]:
                X, y, w, b = make_classification(n_samples=200,
                                                 random_state=10,
                                                 n_classes=n_classes,
                                                 n_informative=n_inf,
                                                 n_features=n_features,
                                                 shared_support=shared_support,
                                                 include_intercept=intercept,
                                                 w_scale=4.)
                mask = np.squeeze(np.logical_not(np.equal(w, 0)))
                for penalty in ['l1', 'l2']:
                    lr = MaskedCoefLogisticRegression(penalty=penalty, C=10.,
                                                      warm_start=True,
                                                      fit_intercept=intercept)
                    lr.fit(X, y, coef_mask=mask)
                    coef_idxs = np.flatnonzero(np.equal(lr.coef_, 0.))
                    coef_idxs = set(coef_idxs.tolist())
                    mask_idxs = np.flatnonzero(np.equal(mask, 0))
                    mask_idxs = set(mask_idxs.tolist())
                    assert mask_idxs.issubset(coef_idxs)
                    lr.fit(X, y, coef_mask=mask)


def test_masked_logistic_standardize():
    """Test the masked logistic regression class with `standardize=True`."""
    n_features = 20
    n_inf = 10
    for shared_support in [True, False]:
        for n_classes in [2, 3]:
            for intercept in [True, False]:
                X, y, w, b = make_classification(n_samples=200,
                                                 random_state=10,
                                                 n_classes=n_classes,
                                                 n_informative=n_inf,
                                                 n_features=n_features,
                                                 shared_support=shared_support,
                                                 include_intercept=intercept,
                                                 w_scale=4.)
                mask = np.squeeze(np.logical_not(np.equal(w, 0)))
                for penalty in ['l1', 'l2']:
                    lr = MaskedCoefLogisticRegression(penalty=penalty, C=10.,
                                                      warm_start=True,
                                                      fit_intercept=intercept,
                                                      standardize=True)
                    lr.fit(X, y, coef_mask=mask)
                    coef_idxs = np.flatnonzero(np.equal(lr.coef_, 0.))
                    coef_idxs = set(coef_idxs.tolist())
                    mask_idxs = np.flatnonzero(np.equal(mask, 0))
                    mask_idxs = set(mask_idxs.tolist())
                    assert mask_idxs.issubset(coef_idxs)
                    lr.fit(X, y, coef_mask=mask)


@pytest.mark.parametrize("n_classes,penalty,fit_intercept", [(3, "l2", True),
                                                             (3, "l2", False),
                                                             (3, "l1", True),
                                                             (3, "l1", False),
                                                             (2, "l2", True),
                                                             (2, "l2", False),
                                                             (2, "l1", True),
                                                             (2, "l1", False)])
def test_masking_with_indexing(n_classes, penalty, fit_intercept):
    """Check that indexing the masks gives the same results as masking with
    logistic regression.
    """
    X, y, w, intercept = make_classification(n_samples=1000,
                                             n_classes=n_classes,
                                             n_features=20,
                                             n_informative=10,
                                             random_state=0)
    mask = w != 0.
    if n_classes == 2:
        mask = mask.ravel()
    coefs, _, _ = _logistic_regression_path(X, y, [10.], coef_mask=mask,
                                            penalty=penalty,
                                            fit_intercept=fit_intercept)
    coefs_old, _, _ = _logistic_regression_path_old(X, y, [10.], coef_mask=mask,
                                                    penalty=penalty,
                                                    fit_intercept=fit_intercept)
    assert_allclose(coefs, coefs_old)
    coefs, _, _ = _logistic_regression_path(X, y, [10.],
                                            penalty=penalty,
                                            fit_intercept=fit_intercept)
    coefs_old, _, _ = _logistic_regression_path_old(X, y, [10.],
                                                    penalty=penalty,
                                                    fit_intercept=fit_intercept)
    assert_allclose(coefs, coefs_old)


@pytest.mark.parametrize("n_classes,penalty,fit_intercept", [(3, "l2", True),
                                                             (3, "l2", False),
                                                             (3, "l1", True),
                                                             (3, "l1", False),
                                                             (2, "l2", True),
                                                             (2, "l2", False),
                                                             (2, "l1", True),
                                                             (2, "l1", False)])
def test_all_masked_with_indexing(n_classes, penalty, fit_intercept):
    """Check masking all of the coef either works with intercept or raises an error.
    """
    X, y, w, intercept = make_classification(n_samples=1000,
                                             n_classes=n_classes,
                                             n_features=20,
                                             n_informative=10,
                                             random_state=0)
    mask = np.zeros_like(w)
    if n_classes == 2:
        mask = mask.ravel()
    coefs, _, _ = _logistic_regression_path(X, y, [10.], coef_mask=mask,
                                            fit_intercept=fit_intercept)
    if fit_intercept:
        if n_classes == 2:
            assert_equal(coefs[0][:-1], 0.)
        else:
            assert_equal(coefs[0][:, :-1], 0.)
    else:
        assert_equal(coefs[0], 0.)


def test_estimation_score_usage():
    """Test the ability to change the estimation score in UoI L1Logistic"""
    methods = ('acc', 'log', 'BIC', 'AIC', 'AICc')
    X, y, w, b = make_classification(n_samples=200,
                                     random_state=6,
                                     n_informative=5,
                                     n_features=10)
    scores = []
    for method in methods:
        l1log = UoI_L1Logistic(random_state=12, estimation_score=method,
                               tol=1e-2, n_boots_sel=24, n_boots_est=24)
        assert_equal(l1log.estimation_score, method)
        l1log.fit(X, y)
        scores.append(l1log.scores_)
    scores = np.stack(scores)
    assert_equal(len(np.unique(scores, axis=0)), len(methods))


def test_set_random_state():
    """Tests whether random states are handled correctly."""
    X, y, w, b = make_classification(n_samples=100,
                                     random_state=60,
                                     n_informative=4,
                                     n_features=5,
                                     w_scale=4.)
    # same state
    l1log_0 = UoI_L1Logistic(random_state=13)
    l1log_1 = UoI_L1Logistic(random_state=13)
    l1log_0.fit(X, y)
    l1log_1.fit(X, y)
    assert_array_equal(l1log_0.coef_, l1log_1.coef_)

    # different state
    l1log_1 = UoI_L1Logistic(random_state=14)
    l1log_1.fit(X, y)
    assert not np.array_equal(l1log_0.coef_, l1log_1.coef_)

    # different state, not set
    l1log_0 = UoI_L1Logistic()
    l1log_1 = UoI_L1Logistic()
    l1log_0.fit(X, y)
    l1log_1.fit(X, y)
    assert not np.array_equal(l1log_0.coef_, l1log_1.coef_)


def test_normalization_by_samples():
    """Test that coef_ does not depend directly on the number of samples."""
    n_features = 20
    for n_classes in [2, 3]:
        X, y, w, b = make_classification(n_samples=200,
                                         random_state=10,
                                         n_classes=n_classes,
                                         n_informative=n_features,
                                         n_features=n_features,
                                         w_scale=4.)
        for penalty in ['l1', 'l2']:
            lr1 = MaskedCoefLogisticRegression(penalty=penalty, C=1e2)
            lr1.fit(X, y)

            lr3 = MaskedCoefLogisticRegression(penalty=penalty, C=1e2)
            lr3.fit(np.tile(X, (3, 1)), np.tile(y, 3))
            assert_allclose(lr1.coef_, lr3.coef_)


def test_l1logistic_binary_strings():
    """Test that binary L1 Logistic runs in the UoI framework."""
    n_inf = 10
    X, y, w, b = make_classification(n_samples=200,
                                     random_state=6,
                                     n_informative=n_inf,
                                     n_features=20,
                                     w_scale=4.,
                                     include_intercept=True)

    classes = ['a', 'b']
    lb = LabelEncoder()
    lb.fit(classes)
    y = lb.inverse_transform(y)

    l1log = UoI_L1Logistic(random_state=10).fit(X, y)
    y_hat = l1log.predict(X)
    assert set(classes) >= set(y_hat)


def test_l1logistic_multiclass_strings():
    """Test that multiclass L1 Logistic runs in the UoI framework when all
       classes share a support."""
    n_features = 20
    n_inf = 10
    X, y, w, b = make_classification(n_samples=200,
                                     random_state=10,
                                     n_classes=5,
                                     n_informative=n_inf,
                                     n_features=n_features,
                                     shared_support=True,
                                     w_scale=4.)
    classes = ['a', 'b', 'c', 'd', 'e']
    lb = LabelEncoder()
    lb.fit(classes)
    y = lb.inverse_transform(y)

    l1log = UoI_L1Logistic(random_state=10).fit(X, y)
    y_hat = l1log.predict(X)
    assert set(classes) >= set(y_hat)


def test_l1logistic_sparse_input():
    """Test that multiclass L1 Logistic works when using sparse matrix
       inputs"""
    rs = np.random.RandomState(17)
    X = sprand(100, 100, random_state=rs)
    classes = ['abc', 'de', 'fgh']
    y = np.array(classes)[rs.randint(3, size=100)]

    kwargs = dict(
        fit_intercept=False,
        random_state=rs,
        n_boots_sel=4,
        n_boots_est=4,
        n_C=7,
    )
    l1log = UoI_L1Logistic(**kwargs).fit(X, y)

    y_hat = l1log.predict(X)
    assert set(classes) >= set(y_hat)


def test_l1logistic_sparse_input_no_center():
    """Test that multiclass L1 Logistic raises an error when asked to center
    sparse data.
    """
    rs = np.random.RandomState(17)
    X = sprand(10, 10, random_state=rs)
    classes = ['abc', 'de', 'fgh']
    y = np.array(classes)[rs.randint(3, size=10)]

    with pytest.raises(ValueError):
        UoI_L1Logistic(fit_intercept=True).fit(X, y)


def test_l1logistic_bad_est_score():
    """Test that multiclass L1 Logistic raises an error when given a bad
    estimation_score value.
    """
    X = np.random.randn(20, 5)
    y = np.ones(20)

    with pytest.raises(ValueError):
        UoI_L1Logistic(estimation_score='z',
                       n_boots_sel=10, n_boots_est=10).fit(X, y)


def test_reg_params():
    """Test whether the upper bound on the regularization parameters correctly
    zero out the coefficients."""
    n_features = 20
    n_inf = 10
    n_classes = 5
    X, y, w, b = make_classification(n_samples=200,
                                     random_state=101,
                                     n_classes=n_classes,
                                     n_informative=n_inf,
                                     n_features=n_features,
                                     shared_support=True)

    uoi_log = UoI_L1Logistic()
    uoi_log.output_dim = n_classes
    reg_params = uoi_log.get_reg_params(X, y)
    C = reg_params[0]['C']
    # check that coefficients get set to zero
    lr = MaskedCoefLogisticRegression(penalty='l1',
                                      C=0.99 * C,
                                      standardize=False,
                                      fit_intercept=True)
    lr.fit(X, y)
    assert_equal(lr.coef_, 0.)

    # check that coefficients above the bound are not set to zero
    lr = MaskedCoefLogisticRegression(penalty='l1',
                                      C=1.01 * C,
                                      standardize=False,
                                      fit_intercept=True)
    lr.fit(X, y)
    assert np.count_nonzero(lr.coef_) > 0


def test_fit_intercept():
    """Tests whether `include_intercept` in passed through to the linear models.
    """
    lr = UoI_L1Logistic(fit_intercept=True)
    assert lr._selection_lm.fit_intercept
    assert lr._estimation_lm.fit_intercept

    lr = UoI_L1Logistic(fit_intercept=False)
    assert not lr._selection_lm.fit_intercept
    assert not lr._estimation_lm.fit_intercept



import abc as _abc
import numpy as np
import logging
from sklearn.linear_model._base import SparseCoefMixin
from sklearn.metrics import r2_score, accuracy_score, log_loss
from sklearn.model_selection import train_test_split
from sklearn.utils import check_X_y
from sklearn.preprocessing import StandardScaler

from scipy.sparse import issparse, csr_matrix

from pyuoi import utils
from pyuoi.mpi_utils import (Gatherv_rows, Bcast_from_root)

from .utils import stability_selection_to_threshold, intersection
from ..utils import check_logger


class AbstractUoILinearModel(SparseCoefMixin, metaclass=_abc.ABCMeta):
    r"""An abstract base class for UoI ``linear_model`` classes.

    Parameters
    ----------
    n_boots_sel : int
        The number of data bootstraps to use in the selection module.
        Increasing this number will make selection more strict.
    n_boots_est : int
        The number of data bootstraps to use in the estimation module.
        Increasing this number will relax selection and decrease variance.
    selection_frac : float
        The fraction of the dataset to use for training in each resampled
        bootstrap, during the selection module. Small values of this parameter
        imply larger "perturbations" to the dataset.
    estimation_frac : float
        The fraction of the dataset to use for training in each resampled
        bootstrap, during the estimation module. The remaining data is used
        to obtain validation scores. Small values of this parameters imply
        larger "perturbations" to the dataset.
    stability_selection : int, float, or array-like
        If int, treated as the number of bootstraps that a feature must appear
        in to guarantee placement in selection profile. If float, must be
        between 0 and 1, and is instead the proportion of bootstraps. If
        array-like, must consist of either ints or floats between 0 and 1.
        In this case, each entry in the array-like object will act as a
        separate threshold for placement in the selection profile.
    fit_intercept : bool
        Whether to calculate the intercept for this model. If set to False,
        no intercept will be used in calculations (e.g. data is expected to be
        already centered).
    standardize : bool
        If True, the regressors X will be standardized before regression by
        subtracting the mean and dividing by their standard deviations.
    shared_support : bool
        For models with more than one output (multinomial logistic regression)
        this determines whether all outputs share the same support or can
        have independent supports.
    max_iter : int
        Maximum number of iterations for iterative fitting methods.
    random_state : int, RandomState instance, or None
        The seed of the pseudo random number generator that selects a random
        feature to update.  If int, random_state is the seed used by the random
        number generator; If RandomState instance, random_state is the random
        number generator; If None, the random number generator is the
        RandomState instance used by ``np.random``.
    comm : MPI communicator
        If passed, the selection and estimation steps are parallelized.
    logger : Logger
        The logger to use for messages when ``verbose=True`` in ``fit``.
        If *None* is passed, a logger that writes to ``sys.stdout`` will be
        used.

    Attributes
    ----------
    coef_ : array, shape (n_features,) or (n_targets, n_features)
        Estimated coefficients for the linear regression problem.
    intercept_ : float
        Independent term in the linear model.
    supports_ : array, shape
        Boolean array indicating whether a given regressor (column) is selected
        for estimation for a given regularization parameter value (row).
    """

    def __init__(self, n_boots_sel=24, n_boots_est=24, selection_frac=0.9,
                 estimation_frac=0.9, stability_selection=1.,
                 fit_intercept=True, standardize=True,
                 shared_support=True, max_iter=None, random_state=None,
                 comm=None, logger=None):
        # data split fractions
        self.selection_frac = selection_frac
        self.estimation_frac = estimation_frac
        # number of bootstraps
        self.n_boots_sel = n_boots_sel
        self.n_boots_est = n_boots_est
        # other hyperparameters
        self.stability_selection = stability_selection
        self.fit_intercept = fit_intercept
        self.standardize = standardize
        self.shared_support = shared_support
        self.max_iter = max_iter
        self.comm = comm
        # preprocessing
        if isinstance(random_state, int):
            # make sure ranks use different seed
            if self.comm is not None:
                random_state += self.comm.rank
            self.random_state = np.random.RandomState(random_state)
        else:
            if random_state is None:
                self.random_state = np.random
            else:
                self.random_state = random_state

        # extract selection thresholds from user provided stability selection
        self.selection_thresholds_ = stability_selection_to_threshold(
            self.stability_selection, self.n_boots_sel)

        self.n_supports_ = None

        self._logger = check_logger(logger, 'uoi_linear_model', self.comm)

    @_abc.abstractproperty
    def estimation_score(self):
        pass

    @_abc.abstractmethod
    def get_reg_params(self):
        pass

    @_abc.abstractstaticmethod
    def _score_predictions(self, metric, fitter, X, y, supports, boot_idxs):
        pass

    @_abc.abstractmethod
    def intersect(self, coef, thresholds):
        """Intersect coefficients across all thresholds."""
        pass

    def _pre_fit(self, X, y):
        """Perform class-specific setup for fit()."""
        if self.standardize:
            if self.fit_intercept and issparse(X):
                msg = ("Cannot center sparse matrices: "
                       "pass `fit_intercept=False`")
                raise ValueError(msg)
            self._X_scaler = StandardScaler(with_mean=self.fit_intercept)
            X = self._X_scaler.fit_transform(X)
        if y.ndim == 2:
            self.output_dim = y.shape[1]
        else:
            self.output_dim = 1
        return X, y

    def _post_fit(self, X, y):
        """Perform class-specific cleanup for fit()."""
        if self.standardize:
            sX = self._X_scaler
            self.coef_ /= sX.scale_[np.newaxis]

    @_abc.abstractmethod
    def _fit_intercept(self, X, y):
        """Fit a model with an intercept and fixed coefficients.

        This is used to re-fit the intercept after the coefficients are
        estimated.
        """
        pass

    @_abc.abstractmethod
    def _fit_intercept_no_features(self, y):
        """Fit a model with only an intercept.

        This is used in cases where the model has no support selected.
        """
        pass

    def fit(self, X, y, stratify=None, verbose=False):
        """Fit data according to the UoI algorithm.

        Parameters
        ----------
        X : ndarray or scipy.sparse matrix, (n_samples, n_features)
            The design matrix.
        y : ndarray, shape (n_samples,)
            Response vector. Will be cast to X's dtype if necessary.
            Currently, this implementation does not handle multiple response
            variables.
        stratify : array-like or None
            Ensures groups of samples are alloted to training/test sets
            proportionally. Labels for each group must be an int greater
            than zero. Must be of size equal to the number of samples, with
            further restrictions on the number of groups.
        verbose : bool
            A switch indicating whether the fitting should print out messages
            displaying progress.
        """
        if verbose:
            self._logger.setLevel(logging.DEBUG)
        else:
            self._logger.setLevel(logging.WARNING)

        X, y = self._pre_fit(X, y)

        X, y = check_X_y(X, y, accept_sparse=['csr', 'csc', 'coo'],
                         y_numeric=True, multi_output=True)

        # extract model dimensions
        n_features = X.shape[1]
        n_coef = self.get_n_coef(X, y)

        # check if the response variable is constant
        if np.unique(y).size == 1:
            self.coef_ = np.zeros((self.output_dim, n_features))
            self._fit_intercept(X, y)
            self._post_fit(X, y)
            return self

        ####################
        # Selection Module #
        ####################
        # choose the regularization parameters for selection sweep
        self.reg_params_ = self.get_reg_params(X, y)
        self.n_reg_params_ = len(self.reg_params_)

        rank = 0
        size = 1
        if self.comm is not None:
            rank = self.comm.rank
            size = self.comm.size

        # initialize selection
        if size > self.n_boots_sel:
            tasks = np.array_split(np.arange(self.n_boots_sel *
                                             self.n_reg_params_), size)[rank]
            selection_coefs = np.empty((tasks.size, n_coef))
            my_boots = dict((task_idx // self.n_reg_params_, None)
                            for task_idx in tasks)
        else:
            # split up bootstraps into processes
            tasks = np.array_split(np.arange(self.n_boots_sel),
                                   size)[rank]
            selection_coefs = np.empty((tasks.size, self.n_reg_params_,
                                        n_coef))
            my_boots = dict((task_idx, None) for task_idx in tasks)

        for boot in range(self.n_boots_sel):
            if size > 1:
                if rank == 0:
                    rvals = train_test_split(np.arange(X.shape[0]),
                                             test_size=1 - self.selection_frac,
                                             stratify=stratify,
                                             random_state=self.random_state)
                else:
                    rvals = [None] * 2
                rvals = [Bcast_from_root(rval, self.comm, root=0)
                         for rval in rvals]
                if boot in my_boots.keys():
                    my_boots[boot] = rvals
            else:
                my_boots[boot] = train_test_split(
                    np.arange(X.shape[0]),
                    test_size=1 - self.selection_frac,
                    stratify=stratify,
                    random_state=self.random_state)

        # iterate over bootstraps
        curr_boot_idx = None
        for ii, task_idx in enumerate(tasks):
            if size > self.n_boots_sel:
                boot_idx = task_idx // self.n_reg_params_
                reg_idx = task_idx % self.n_reg_params_
                my_reg_params = [self.reg_params_[reg_idx]]
            else:
                boot_idx = task_idx
                my_reg_params = self.reg_params_
            # Never warm start across bootstraps
            if (curr_boot_idx != boot_idx):
                if hasattr(self._selection_lm, 'coef_'):
                    self._selection_lm.coef_ *= 0.
                if hasattr(self._selection_lm, 'intercept_'):
                    self._selection_lm.intercept_ *= 0.
            curr_boot_idx = boot_idx

            # draw a resampled bootstrap
            idxs_train, idxs_test = my_boots[boot_idx]
            X_rep = X[idxs_train]
            y_rep = y[idxs_train]

            # fit the coefficients
            if size > self.n_boots_sel:
                msg = ("selection bootstrap %d, "
                       "regularization parameter set %d"
                       % (boot_idx, reg_idx))
                self._logger.info(msg)

            else:
                self._logger.info("selection bootstrap %d" % (boot_idx))
            selection_coefs[ii] = np.squeeze(
                self.uoi_selection_sweep(X_rep, y_rep, my_reg_params))

        # if distributed, gather selection coefficients to 0,
        # perform intersection, and broadcast results
        if size > 1:
            selection_coefs = Gatherv_rows(selection_coefs, self.comm, root=0)
            if rank == 0:
                if size > self.n_boots_sel:
                    selection_coefs = selection_coefs.reshape(
                        self.n_boots_sel,
                        self.n_reg_params_,
                        n_coef)
                supports = self.intersect(
                    selection_coefs,
                    self.selection_thresholds_).astype(int)
            else:
                supports = None
            supports = Bcast_from_root(supports, self.comm, root=0)
            self.supports_ = supports.astype(bool)
        else:
            self.supports_ = self.intersect(selection_coefs,
                                            self.selection_thresholds_)

        self.n_supports_ = self.supports_.shape[0]

        if rank == 0:
            self._logger.info("Found %d supports" % self.n_supports_)

        #####################
        # Estimation Module #
        #####################
        # set up data arrays
        tasks = np.array_split(np.arange(self.n_boots_est *
                                         self.n_supports_), size)[rank]
        my_boots = dict((task_idx // self.n_supports_, None)
                        for task_idx in tasks)
        estimates = np.zeros((tasks.size, n_coef))

        for boot in range(self.n_boots_est):
            if size > 1:
                if rank == 0:
                    rvals = train_test_split(np.arange(X.shape[0]),
                                             test_size=1 - self.estimation_frac,
                                             stratify=stratify,
                                             random_state=self.random_state)
                else:
                    rvals = [None] * 2
                rvals = [Bcast_from_root(rval, self.comm, root=0)
                         for rval in rvals]
                if boot in my_boots.keys():
                    my_boots[boot] = rvals
            else:
                my_boots[boot] = train_test_split(
                    np.arange(X.shape[0]),
                    test_size=1 - self.estimation_frac,
                    stratify=stratify,
                    random_state=self.random_state)

        # score (r2/AIC/AICc/BIC) for each bootstrap for each support
        scores = np.zeros(tasks.size)

        # iterate over bootstrap samples and supports
        for ii, task_idx in enumerate(tasks):
            boot_idx = task_idx // self.n_supports_
            support_idx = task_idx % self.n_supports_
            support = self.supports_[support_idx]
            # draw a resampled bootstrap
            idxs_train, idxs_test = my_boots[boot_idx]
            X_rep = X[idxs_train]
            y_rep = y[idxs_train]
            self._logger.info("estimation bootstrap %d, support %d"
                              % (boot_idx, support_idx))
            if np.any(support):

                # compute the estimate and store the fitted coefficients
                if self.shared_support:
                    self._estimation_lm.fit(X_rep[:, support], y_rep)
                    estimates[ii, np.tile(support, self.output_dim)] = \
                        self._estimation_lm.coef_.ravel()
                else:
                    self._estimation_lm.fit(X_rep, y_rep, coef_mask=support)
                    estimates[ii] = self._estimation_lm.coef_.ravel()

                scores[ii] = self._score_predictions(
                    metric=self.estimation_score,
                    fitter=self._estimation_lm,
                    X=X, y=y,
                    support=support,
                    boot_idxs=my_boots[boot_idx])
            else:
                fitter = self._fit_intercept_no_features(y_rep)
                if issparse(X):
                    X_ = csr_matrix(X.shape, dtype=X.dtype)
                else:
                    X_ = np.zeros_like(X)
                scores[ii] = self._score_predictions(
                    metric=self.estimation_score,
                    fitter=fitter,
                    X=X_, y=y,
                    support=np.zeros(X.shape[1], dtype=bool),
                    boot_idxs=my_boots[boot_idx])

        if size > 1:
            estimates = Gatherv_rows(send=estimates, comm=self.comm,
                                     root=0)
            scores = Gatherv_rows(send=scores, comm=self.comm,
                                  root=0)
            self.rp_max_idx_ = None
            best_estimates = None
            coef = None
            self.intercept_ = None
            if rank == 0:
                estimates = estimates.reshape(self.n_boots_est,
                                              self.n_supports_, n_coef)
                scores = scores.reshape(self.n_boots_est, self.n_supports_)
                self.rp_max_idx_ = np.argmax(scores, axis=1)
                best_estimates = estimates[np.arange(self.n_boots_est),
                                           self.rp_max_idx_]
                # take the median across estimates for the final estimate
                coef = np.median(best_estimates,
                                 axis=0).reshape(self.output_dim, n_features)
                self.coef_ = coef
                self._fit_intercept(X, y)
            self.estimates_ = Bcast_from_root(estimates, self.comm, root=0)
            self.scores_ = Bcast_from_root(scores, self.comm, root=0)
            self.coef_ = Bcast_from_root(coef, self.comm, root=0)
            self.intercept_ = Bcast_from_root(self.intercept_,
                                              self.comm, root=0)
            self.rp_max_idx_ = self.comm.bcast(self.rp_max_idx_, root=0)
        else:
            self.estimates_ = estimates.reshape(self.n_boots_est,
                                                self.n_supports_, n_coef)
            self.scores_ = scores.reshape(self.n_boots_est, self.n_supports_)
            self.rp_max_idx_ = np.argmax(self.scores_, axis=1)
            # extract the estimates over bootstraps from model with best
            # regularization parameter value
            best_estimates = self.estimates_[np.arange(self.n_boots_est),
                                             self.rp_max_idx_, :]
            # take the median across estimates for the final, bagged estimate
            self.coef_ = np.median(best_estimates,
                                   axis=0).reshape(self.output_dim, n_features)
            self._fit_intercept(X, y)
        self._post_fit(X, y)

        return self

    def uoi_selection_sweep(self, X, y, reg_param_values):
        """Perform selection regression on a dataset over a sweep of
        regularization parameter values.

        Parameters
        ----------
        X : ndarray or scipy.sparse matrix, shape (n_samples, n_features)
            The design matrix.
        y : ndarray, shape (n_samples,)
            Response vector.
        reg_param_values: list of dicts
            A list of dictionaries containing the regularization parameter
            values to iterate over.

        Returns
        -------
        coefs : ndarray, shape (n_param_values, n_features)
            Predicted parameter values for each regularization strength.
        """

        n_param_values = len(reg_param_values)
        n_coef = self.get_n_coef(X, y)

        coefs = np.zeros((n_param_values, n_coef))

        # apply the selection regression to bootstrapped datasets
        for reg_param_idx, reg_params in enumerate(reg_param_values):
            # reset the regularization parameter
            self._selection_lm.set_params(**reg_params)
            # rerun fit
            self._selection_lm.fit(X, y)
            # store coefficients
            coefs[reg_param_idx] = self._selection_lm.coef_.ravel()

        return coefs

    def get_n_coef(self, X, y):
        """Return the number of coefficients that will be estimated

        This should return the shape of X.
        """
        return X.shape[1] * self.output_dim


class AbstractUoILinearRegressor(AbstractUoILinearModel,
                                 metaclass=_abc.ABCMeta):
    """An abstract base class for UoI linear regression classes."""

    _valid_estimation_metrics = ('r2', 'AIC', 'AICc', 'BIC')

    _train_test_map = {'train': 0, 'test': 1}

    _default_est_targets = {'r2': 1, 'AIC': 0, 'AICc': 0, 'BIC': 0}

    def __init__(self, n_boots_sel=24, n_boots_est=24, selection_frac=0.9,
                 estimation_frac=0.9, stability_selection=1.,
                 estimation_score='r2', estimation_target=None,
                 copy_X=True, fit_intercept=True,
                 standardize=True, random_state=None, max_iter=None,
                 comm=None, logger=None):
        super(AbstractUoILinearRegressor, self).__init__(
            n_boots_sel=n_boots_sel,
            n_boots_est=n_boots_est,
            selection_frac=selection_frac,
            estimation_frac=estimation_frac,
            stability_selection=stability_selection,
            fit_intercept=fit_intercept,
            standardize=standardize,
            max_iter=max_iter,
            random_state=random_state,
            comm=comm,
            logger=logger)

        if estimation_score not in self._valid_estimation_metrics:
            raise ValueError(
                "invalid estimation metric: '%s'" % estimation_score)

        self.__estimation_score = estimation_score

        if estimation_target is not None:
            if estimation_target not in ['train', 'test']:
                raise ValueError(
                    "invalid estimation target: %s" % estimation_target)
            else:
                estimation_target = self._train_test_map[estimation_target]
        else:
            estimation_target = self._default_est_targets[estimation_score]
        self._estimation_target = estimation_target

    def _pre_fit(self, X, y):
        X, y = super()._pre_fit(X, y)
        if y.ndim == 1:
            y = y[:, np.newaxis]
        elif y.ndim == 2:
            if y.shape[1] > 1:
                raise ValueError('y should either have shape ' +
                                 '(n_samples, ) or (n_samples, 1).')
        else:
            raise ValueError('y should either have shape ' +
                             '(n_samples, ) or (n_samples, 1).')
        if self.standardize:
            self._y_scaler = StandardScaler(with_mean=self.fit_intercept)
            y = self._y_scaler.fit_transform(y)
        y = np.squeeze(y)
        self.output_dim = 1
        return X, y

    def _post_fit(self, X, y):
        super()._post_fit(X, y)
        if self.standardize:
            sX = self._X_scaler
            sy = self._y_scaler
            self.coef_ *= sy.scale_[:, np.newaxis]
            if self.fit_intercept:
                self.intercept_ *= sy.scale_
                self.intercept_ += sy.mean_ - np.dot(sX.mean_,
                                                     self.coef_.T)
        self.coef_ = np.squeeze(self.coef_)

    def intersect(self, coef, thresholds):
        """Intersect coefficients accross all thresholds."""
        return intersection(coef, thresholds)

    @property
    def estimation_score(self):
        return self.__estimation_score

    def _score_predictions(self, metric, fitter, X, y, support, boot_idxs):
        """Score, according to some metric, predictions provided by a model.

        The resulting score will be negated if an information criterion is
        specified.

        Parameters
        ----------
        metric : string
            The type of score to run on the prediction. Valid options include
            'r2' (explained variance), 'BIC' (Bayesian information criterion),
            'AIC' (Akaike information criterion), and 'AICc' (corrected AIC).
        fitter : object
            Must contain .predict and .predict_proba methods.
        X : array-like
            The design matrix.
        y : array-like
            Response vector.
        supports : array-like
            The value of the supports for the model
        boot_idxs : 2-tuple of array-like objects
            Tuple of (train_idxs, test_idxs) generated from a bootstrap
            sample. If this is specified, then the appropriate set of
            data will be used for evaluating scores: test data for r^2,
            and training data for information criteria

        Returns
        -------
        score : float
            The score.
        """

        # Select the data relevant for the estimation_score
        X = X[boot_idxs[self._estimation_target]]
        y = y[boot_idxs[self._estimation_target]]

        if y.ndim == 2:
            if y.shape[1] > 1:
                raise ValueError('y should either have shape ' +
                                 '(n_samples, ) or (n_samples, 1).')
            y = np.squeeze(y)
        elif y.ndim > 2:
            raise ValueError('y should either have shape ' +
                             '(n_samples, ) or (n_samples, 1).')

        y_pred = fitter.predict(X[:, support])
        if y.shape != y_pred.shape:
            raise ValueError('Targets and predictions are not the same shape.')

        if metric == 'r2':
            score = r2_score(y, y_pred)
        else:
            ll = utils.log_likelihood_glm(model='normal',
                                          y_true=y,
                                          y_pred=y_pred)
            n_features = np.count_nonzero(support)
            n_samples = X.shape[0]
            if metric == 'BIC':
                score = utils.BIC(ll, n_features, n_samples)
            elif metric == 'AIC':
                score = utils.AIC(ll, n_features)
            elif metric == 'AICc':
                score = utils.AICc(ll, n_features, n_samples)
            else:
                raise ValueError(metric + ' is not a valid option.')
            # negate the score since lower information criterion is preferable
            score = -score
        return score

    def _fit_intercept_no_features(self, y):
        """Fit a model with only an intercept.

        This is used in cases where the model has no support selected.
        """
        return LinearInterceptFitterNoFeatures(y)

    def _fit_intercept(self, X, y):
        """Fit the intercept."""
        if self.fit_intercept:
            self.intercept_ = (y.mean(axis=0) -
                               np.dot(X.mean(axis=0), self.coef_.T))
        else:
            self.intercept_ = np.zeros(1)


class LinearInterceptFitterNoFeatures(object):
    def __init__(self, y):
        self.intercept_ = y.mean()

    def predict(self, X):
        n_samples = X.shape[0]
        return np.tile(self.intercept_, n_samples)


class AbstractUoIGeneralizedLinearRegressor(AbstractUoILinearModel,
                                            metaclass=_abc.ABCMeta):
    """An abstract base class for UoI linear classifier classes."""

    _valid_estimation_metrics = ('log', 'BIC', 'AIC', 'AICc', 'acc')

    _train_test_map = {'train': 0, 'test': 1}

    _default_est_targets = {'log': 1, 'AIC': 0, 'AICc': 0,
                            'BIC': 0, 'acc': 1}

    def __init__(self, n_boots_sel=24, n_boots_est=24, selection_frac=0.9,
                 estimation_frac=0.9, stability_selection=1.,
                 estimation_score='acc', estimation_target=None,
                 copy_X=True, fit_intercept=True, standardize=True,
                 random_state=None, max_iter=None, shared_support=True,
                 comm=None, logger=None):
        super(AbstractUoIGeneralizedLinearRegressor, self).__init__(
            n_boots_sel=n_boots_sel,
            n_boots_est=n_boots_est,
            selection_frac=selection_frac,
            estimation_frac=estimation_frac,
            stability_selection=stability_selection,
            random_state=random_state,
            fit_intercept=fit_intercept,
            standardize=standardize,
            shared_support=shared_support,
            max_iter=max_iter,
            comm=comm,
            logger=logger)

        if estimation_score not in self._valid_estimation_metrics:
            raise ValueError(
                "invalid estimation metric: '%s'" % estimation_score)
        self.__estimation_score = estimation_score

        if estimation_target is not None:
            if estimation_target not in ['train', 'test']:
                raise ValueError(
                    "invalid estimation target: %s" % estimation_target)
            else:
                estimation_target = self._train_test_map[estimation_target]
        else:
            estimation_target = self._default_est_targets[estimation_score]

        self._estimation_target = estimation_target

    def _post_fit(self, X, y):
        super()._post_fit(X, y)
        if self.standardize and self.fit_intercept:
            sX = self._X_scaler
            self.intercept_ += np.dot(sX.mean_ * sX.scale_,
                                      self.coef_.T)

    def intersect(self, coef, thresholds):
        """Intersect coefficients accross all thresholds.

        This implementation will account for multi-class classification.
        """
        supports = intersection(coef, thresholds)
        if self.output_dim > 1 and self.shared_support:
            n_features = supports.shape[-1] // self.output_dim
            supports = supports.reshape((-1, self.output_dim, n_features))
            supports = np.sum(supports, axis=-2).astype(bool)
            supports = np.unique(supports, axis=0)
        return supports

    @property
    def estimation_score(self):
        return self.__estimation_score

    def _score_predictions(self, metric, fitter, X, y, support, boot_idxs):
        """Score, according to some metric, predictions provided by a model.

        The resulting score will be negated if an information criterion is
        specified.

        Parameters
        ----------
        metric : string
            The type of score to run on the prediction. Valid options include
            'r2' (explained variance), 'BIC' (Bayesian information criterion),
            'AIC' (Akaike information criterion), and 'AICc' (corrected AIC).
        fitter : object
            Must contain .predict and .predict_proba methods.
        X : array-like
            The design matrix.
        y : array-like
            Response vector.
        supports : array-like
            The value of the supports for the model
        boot_idxs : 2-tuple of array-like objects
            Tuple of (train_idxs, test_idxs) generated from a bootstrap
            sample. If this is specified, then the appropriate set of
            data will be used for evaluating scores: test data for r^2,
            and training data for information criteria

        Returns
        -------
        score : float
            The score.
        """

        # Select the data relevant for the estimation_score
        X = X[boot_idxs[self._estimation_target]]
        y = y[boot_idxs[self._estimation_target]]

        if metric == 'acc':
            if self.shared_support:
                y_pred = fitter.predict(X[:, support])
            else:
                y_pred = fitter.predict(X)
            score = accuracy_score(y, y_pred)
        else:

            if self.shared_support:
                y_pred = fitter.predict_proba(X[:, support])
            else:
                y_pred = fitter.predict_proba(X)
            ll = -log_loss(y, y_pred, labels=self.classes_)
            if metric == 'log':
                score = ll
            else:
                n_features = np.count_nonzero(support)
                n_samples = X.shape[0]
                if metric == 'BIC':
                    score = utils.BIC(n_samples * ll, n_features, n_samples)
                elif metric == 'AIC':
                    score = utils.AIC(n_samples * ll, n_features)
                elif metric == 'AICc':
                    score = utils.AICc(n_samples * ll, n_features, n_samples)
                else:
                    raise ValueError(metric + ' is not a valid metric.')
                # negate the score since lower information criterion is
                # preferable
                score = -score

        return score



import scipy.optimize as spo
import numpy as np
import logging

from .base import AbstractDecompositionModel
from .utils import dissimilarity

from ..utils import check_logger

from itertools import combinations
from sklearn.base import BaseEstimator
from sklearn.decomposition import NMF as skNMF
from sklearn.cluster import DBSCAN
from sklearn.preprocessing import normalize
from sklearn.utils.validation import check_non_negative
from sklearn.utils.validation import check_is_fitted


class UoI_NMF_Base(AbstractDecompositionModel):
    """Performs non-negative matrix factorization in the Union of Intersections
    framework.

    This base class accepts objects or functions that perform the NMF fitting,
    clustering, non-negative regression, and consensus grouping.

    Parameters
    ----------
    n_boots : int
        The number of bootstraps to use for model selection.
    ranks : int, list, or None
        The range of k to use. If *ranks* is an int, range(2, ranks + 1) will
        be used. If not specified, range(X.shape[1]) will be used.
    nmf : NMF object
        The NMF object to use to perform fitting.
        Note: this class must take *n_components* as an argument.
    cluster : Clustering object
        Clustering object to use. If None, defaults to DBSCAN.
    nnreg : NNLS object
        Non-negative regressor to use. If None, defaults to
        scipy.optimize.nnls.
    cons_meth : function
        The method for computing consensus bases after clustering. If None,
        uses np.mean.
    use_dissimilarity : bool
        Whether to use dissimilarity to choose the final rank. If False, all
        bases across ranks are concatenated and clustered. The final rank in
        this case is how many clusters are chosen.
    random_state : int, RandomState instance, or None
        The seed of the pseudo random number generator that selects a random
        feature to update.  If int, random_state is the seed used by the random
        number generator; If RandomState instance, random_state is the random
        number generator; If None, the random number generator is the
        RandomState instance used by `np.random`.
    logger : Logger
        The logger to use for messages when ``verbose=True`` in ``fit``.
        If *None* is passed, a logger that writes to ``sys.stdout`` will be
        used.
    """
    def __init__(
        self, n_boots=10, ranks=None, nmf=None, cluster=None, nnreg=None,
        cons_meth=None, use_dissimilarity=True, random_state=None, logger=None
    ):
        super(UoI_NMF_Base, self).__init__()
        self.__initialize(
            n_boots=n_boots,
            ranks=ranks,
            nmf=nmf,
            cluster=cluster,
            nnreg=nnreg,
            cons_meth=cons_meth,
            logger=logger,
            random_state=random_state)
        if self.n_boots == 1 and use_dissimilarity:
            raise ValueError("Cannot use dissimilarity to compute best rank "
                             "with a single bootstrap")
        self.use_dissimilarity = use_dissimilarity

    def set_params(self, **kwargs):
        """Set the parameters of this estimator."""
        super(UoI_NMF_Base, self).set_params(**kwargs)
        self.__initialize(**self.get_params())

    def __initialize(self, **kwargs):
        """Initializes the NMF class."""
        n_boots = kwargs['n_boots']
        ranks = kwargs['ranks']
        nmf = kwargs['nmf']
        cluster = kwargs['cluster']
        nnreg = kwargs['nnreg']
        cons_meth = kwargs['cons_meth']
        random_state = kwargs['random_state']

        self.n_boots = n_boots
        self.components_ = None
        logger = kwargs['logger']

        # initialize NMF ranks to use
        if ranks is not None:
            if isinstance(ranks, int):
                self.ranks = list(range(2, ranks + 1)) \
                    if isinstance(ranks, int) else list(ranks)
            elif isinstance(ranks, (list, tuple, range, np.ndarray)):
                self.ranks = tuple(ranks)
            else:
                raise ValueError('Specify a max value or an array-like for k.')

        # initialize NMF solver
        if nmf is not None:
            if isinstance(nmf, type):
                raise ValueError('nmf must be an instance, not a class.')
            self.nmf = nmf
        else:
            self.nmf = skNMF(beta_loss='kullback-leibler', solver='mu',
                             max_iter=400, init='random')

        # initialize clusterer
        if cluster is not None:
            if isinstance(cluster, type):
                raise ValueError('dbscan must be an instance, not a class.')
            self.cluster = cluster
        else:
            self.cluster = DBSCAN(min_samples=self.n_boots / 2)

        # initialize non-negative regression solver
        if nnreg is None:
            self.nnreg = lambda A, b: spo.nnls(A, b)[0]
        else:
            if isinstance(nnreg, BaseEstimator):
                self.nnreg = lambda A, B: nnreg.fit(A, B).coef_
            elif callable(nnreg):
                self.nnreg = nnreg
            else:
                raise ValueError("Unrecognized regressor.")

        # initialize method for computing consensus H bases after clustering
        if cons_meth is None:
            # default uses mean
            self.cons_meth = np.mean
        else:
            self.cons_meth = cons_meth

        # initialize random state
        if random_state is None:
            self._rand = np.random
        else:
            if isinstance(random_state, int):
                self._rand = np.random.RandomState(random_state)
            elif isinstance(random_state, np.random.RandomState):
                self._rand = random_state
            self.nmf.set_params(random_state=self._rand)

        self.components_ = None
        self.bases_samples_ = None
        self.bases_samples_labels_ = None
        self.bootstraps_ = None

        self.comm = None

        self._logger = check_logger(logger, 'uoi_decomposition', self.comm)

    def fit(self, X, verbose=False):
        r"""Compute the basis matrix on the provided data matrix using the
        UoI\ :sub:`NMF` algorithm.

        Parameters
        ----------
        X : ndarray, shape (n_samples, n_features)
            Data matrix to be decomposed.
        verbose : bool
            If True, outputs status updates.
        """
        if verbose:
            self._logger.setLevel(logging.DEBUG)
        else:
            self._logger.setLevel(logging.WARNING)

        check_non_negative(X, 'UoI NMF')
        n_samples, n_features = X.shape

        H_samples = {k: np.zeros((self.n_boots, k, n_features))
                     for k in self.ranks}

        rep_idx = self._rand.randint(n_samples, size=(self.n_boots, n_samples))
        for boot_idx in range(self.n_boots):
            self._logger.info("Bootstrap %d" % boot_idx)
            # compute NMF bases for k across bootstrap replicates
            sample = X[rep_idx[boot_idx]]
            for k_idx, k in enumerate(self.ranks):
                # concatenate k by p
                H_samples[k][boot_idx] = \
                    self.nmf.set_params(n_components=k).fit(sample).components_

        if self.use_dissimilarity:
            gamma = np.zeros(len(self.ranks))
            # iterate over each rank
            for k_idx, k in enumerate(self.ranks):
                # extract the bases for each rank
                H_k = H_samples[k]
                for boot1, boot2 in combinations(range(self.n_boots), 2):
                    gamma[k_idx] += dissimilarity(H_k[boot1], H_k[boot2])
            self.dissimilarity_ = (gamma * 2) /\
                                  (self.n_boots *
                                   (self.n_boots - 1))
            k_min = self.ranks[np.argmin(gamma)]
            H_pre_cluster = H_samples[k_min].reshape((self.n_boots * k_min,
                                                      n_features))
        else:
            H_pre_cluster = np.zeros((self.n_boots * np.sum(self.ranks),
                                      n_features))
            start_idx = 0
            for k in self.ranks:
                H_k = H_samples[k].reshape((self.n_boots * k, n_features))
                end_idx = start_idx + self.n_boots * k
                H_pre_cluster[start_idx:end_idx] = H_k
                start_idx = end_idx

        # remove zero bases and normalize across features
        H_pre_cluster = H_pre_cluster[np.sum(H_pre_cluster, axis=1) != 0.0]
        H_pre_cluster = normalize(H_pre_cluster, norm='l2', axis=1)

        # cluster all bases
        self._logger.info("clustering bases samples")
        labels = self.cluster.fit_predict(H_pre_cluster)

        # compute consensus bases from clusters
        cluster_ids = np.unique(labels[labels != -1])
        n_clusters = cluster_ids.size
        self._logger.info("found %d bases, computing consensus bases" %
                          n_clusters)
        H_cons = np.zeros((n_clusters, n_features))

        for c_id in cluster_ids:
            H_cons[c_id, :] = self.cons_meth(H_pre_cluster[labels == c_id],
                                             axis=0)

        # re-normalize across features
        H_cons = normalize(H_cons, norm='l2', axis=1)

        self.components_ = H_cons
        self.n_components = self.components_.shape[0]
        self.bases_samples_ = H_pre_cluster
        self.bases_samples_labels_ = labels
        self.bootstraps_ = rep_idx
        self.reconstruction_err_ = None
        return self

    def transform(self, X, reconstruction_err=True):
        r"""Transform the data according to the fitted UoI\ :sub:`NMF` model.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Data matrix to be decomposed.
        reconstruction_err : bool
            If True, the reconstruction error is computed and stored as a class
            attribute.

        Returns
        -------
        W : array-like, shape (n_samples, n_components)
            Transformed data (coefficients of bases).
        """
        check_is_fitted(self, ['components_'])
        n_samples, n_features = X.shape

        if n_features != self.components_.shape[1]:
            raise ValueError(
                'Incompatible shape: cannot reconstruct with %s and %s'
                % (X.shape, self.components_.shape))

        H_t = self.components_.T
        W = np.zeros((n_samples, self.n_components), dtype=X.dtype)

        # calculate coefficients for each sample
        for sample in range(n_samples):
            W[sample] = self.nnreg(H_t, X[sample])

        # calculate reconstruction error
        if reconstruction_err:
            self.reconstruction_err_ = np.linalg.norm(
                X - self.inverse_transform(W))

        return W

    def fit_transform(self, X, reconstruction_err=True, verbose=None):
        r"""Fit and transform the data according to the fitted UoI\ :sub:`NMF`
        model.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Data matrix to be decomposed.
        reconstruction_err : bool
            If True, the reconstruction error is computed and stored as a class
            attribute.
        verbose : bool
            If True, outputs status updates.

        Returns
        -------
        W : array-like, shape (n_samples, n_components)
            Transformed data (coefficients of bases).
        """
        self.fit(X, verbose=verbose)
        return self.transform(X, reconstruction_err=reconstruction_err)

    def inverse_transform(self, W):
        """Transform data back to its original space.

        Parameters
        ----------
        W : array-like, shape (n_samples, n_components)
            Transformed data matrix.

        Returns
        -------
        X : array-like, shape (n_samples, n_features)
            Data matrix of original shape.
        """
        check_is_fitted(self, ['components_'])
        n_components = W.shape[1]

        if n_components != self.components_.shape[0]:
            raise ValueError(
                'Incompatible shape: cannot multiply %s with %s.'
                % (W.shape, self.components_.shape))

        return np.matmul(W, self.components_)


class UoI_NMF(UoI_NMF_Base):
    r"""Performs non-negative matrix factorization in the Union of
    Intersections framework.

    This derived class uses (and accepts the keyword arguments for)
    ``scikit-learn``’s NMF and DBSCAN objects, ``scipy``’s non-negative least
    squares function, and a mean function for consensus grouping.

    Parameters
    ----------
    n_boots : int
        The number of bootstraps to use for model selection.
    ranks : int, list, or None
        The range of k to use. If *ranks* is an int, range(2, ranks + 1) will
        be used. If not specified, range(X.shape[1]) will be used.
    nmf_init :  "random" | "nndsvd" |  "nndsvda" | "nndsvdar" | "custom"
        Method used to initialize the NMF procedure. Valid options:
            * "random": non-negative random matrices, scaled with
                :code:`sqrt(X.mean() / n_components)`
            * "nndsvd": Nonnegative Double Singular Value Decomposition (NNDSVD)
                initialization (better for sparseness)
            * "nndsvda": NNDSVD with zeros filled with the average of X
                (better when sparsity is not desired)
            * "nndsvdar": NNDSVD with zeros filled with small random values
                (generally faster, less accurate alternative to NNDSVDa
                for when sparsity is not desired)
            * "custom": use custom matrices W and H
    nmf_solver : 'cd' | 'mu', optional
        Numerical solver to use for NMF: 'cd' is a Coordinate Descent solver,
        while 'mu' is a Multiplicative Update solver.
    nmf_beta_loss : float or string, optional
        String must be in {'frobenius', 'kullback-leibler', 'itakura-saito'}.
        Beta divergence to be minimized, measuring the distance between X
        and the dot product WH. Note that values different from 'frobenius'
        (or 2) and 'kullback-leibler' (or 1) lead to significantly slower
        fits. Note that for beta_loss <= 0 (or 'itakura-saito'), the input
        matrix X cannot contain zeros. Used only in 'mu' solver.
    nmf_tol : float, optional
        Tolerance of the stopping condition for NMF algorithm.
    nmf_max_iter : integer, optional
        Maximum number of iterations before timing out in NMF.
    db_eps : float, optional
        The maximum distance between two samples for them to be considered
        as in the same neighborhood in the DBSCAN algorithm.
    db_min_samples : int, optional
        The number of samples (or total weight) in a neighborhood for a point
        to be considered as a core point. This includes the point itself.
    db_metric : string, or callable, optional
        The metric to use when calculating distance between instances in a
        feature array. If metric is a string or callable, it must be one of
        the options allowed by :func:`sklearn.metrics.pairwise_distances` for
        its metric parameter.
        If metric is "precomputed", X is assumed to be a distance matrix and
        must be square. X may be a sparse matrix, in which case only "nonzero"
        elements may be considered neighbors for DBSCAN.
    db_metric_params : dict, optional
        Additional keyword arguments for the metric function.
    db_algorithm : {'auto', 'ball_tree', 'kd_tree', 'brute'}, optional
        The algorithm to be used by the NearestNeighbors module
        to compute pointwise distances and find nearest neighbors.
        See NearestNeighbors module documentation for details.
    db_leaf_size : int, optional (default = 30)
        Leaf size passed to BallTree or cKDTree. This can affect the speed
        of the construction and query, as well as the memory required
        to store the tree. The optimal value depends
        on the nature of the problem.
    random_state : int, RandomState instance, or None
        The seed of the pseudo random number generator that selects a random
        feature to update.  If int, random_state is the seed used by the random
        number generator; If RandomState instance, random_state is the random
        number generator; If None, the random number generator is the
        RandomState instance used by `np.random`.
    logger : Logger
        The logger to use for messages when ``verbose=True`` in ``fit``.
        If *None* is passed, a logger that writes to ``sys.stdout`` will be
        used.
    """
    def __init__(
        self, n_boots, ranks=None,
        nmf_init='random', nmf_solver='mu', nmf_beta_loss='kullback-leibler',
        nmf_tol=0.0001, nmf_max_iter=400,
        db_eps=0.5, db_min_samples=None, db_metric='euclidean',
        db_metric_params=None, db_algorithm='auto', db_leaf_size=30,
        use_dissimilarity=True, random_state=None, logger=None,
        nmf=None, cluster=None, nnreg=None, cons_meth=None
    ):
        # create NMF solver
        nmf = skNMF(init=nmf_init,
                    solver=nmf_solver,
                    beta_loss=nmf_beta_loss,
                    tol=nmf_tol,
                    max_iter=nmf_max_iter)

        # create DBSCAN solver
        if db_min_samples is None:
            db_min_samples = n_boots / 2
        dbscan = DBSCAN(eps=db_eps,
                        min_samples=db_min_samples,
                        metric=db_metric,
                        metric_params=db_metric_params,
                        algorithm=db_algorithm)

        super(UoI_NMF, self).__init__(
            n_boots=n_boots,
            ranks=ranks,
            nmf=nmf,
            cluster=dbscan,
            nnreg=None,
            cons_meth=np.mean,
            random_state=random_state,
            use_dissimilarity=use_dissimilarity,
            logger=logger)



import pytest
from numpy.testing import assert_array_equal, assert_array_almost_equal

from pyuoi.lbfgs import LBFGS, fmin_lbfgs
from pyuoi.lbfgs._lowlevel import LBFGSError
import numpy as np


def test_fmin_lbfgs():
    def f(x, g, *args):
        g[0] = 2 * x
        return x ** 2

    xmin = fmin_lbfgs(f, 100., line_search='armijo')
    assert_array_equal(xmin, [0])

    xmin = fmin_lbfgs(f, 100., line_search='strongwolfe')
    assert_array_equal(xmin, [0])


class TestOWLQN:

    def test_owl_qn_end(self):
        def f(x, g, *args):
            g[:] = 2. * (x - 1.)
            return np.sum((x - 1.) ** 2)

        xmin = fmin_lbfgs(f, np.zeros(10), orthantwise_c=1.,
                          orthantwise_end=5)
        assert_array_equal(xmin[5:], 1.)
        assert np.all(xmin[:5] < 1.)

    def test_owl_qn(self):
        def f(x, g, *args):
            g[0] = 2 * x
            return x ** 2

        xmin = fmin_lbfgs(f, 100., orthantwise_c=1, line_search='wolfe')
        assert_array_equal(xmin, [0])

    def test_owl_line_search_warning_explicit(self):
        def f(x, g, *args):
            g[0] = 2 * x
            return x ** 2

        with pytest.warns(UserWarning, match="OWL-QN"):
            fmin_lbfgs(f, 100., orthantwise_c=1, line_search='morethuente')
        with pytest.warns(UserWarning, match="OWL-QN"):
            fmin_lbfgs(f, 100., orthantwise_c=1, line_search='armijo')
        with pytest.warns(UserWarning, match="OWL-QN"):
            fmin_lbfgs(f, 100., orthantwise_c=1, line_search='strongwolfe')

    @pytest.mark.xfail(strict=True)
    def test_owl_wolfe_no_warning(self):
        """ This test is an attempt to show that wolfe throws no warnings.
        """

        def f(x, g, *args):
            g[0] = 2 * x
            return x ** 2

        with pytest.warns(UserWarning, match="OWL-QN"):
            fmin_lbfgs(f, 100., orthantwise_c=1, line_search='wolfe')


def test_2d():
    def f(x, g, f_calls):
        assert x.shape == (2, 2)
        assert g.shape == x.shape
        g[:] = 2 * x
        f_calls[0] += 1
        return (x ** 2).sum()

    def progress(x, g, fx, xnorm, gnorm, step, k, ls, *args):
        assert x.shape == (2, 2)
        assert g.shape == x.shape

        assert np.sqrt((x ** 2).sum()) == xnorm
        assert np.sqrt((g ** 2).sum()) == gnorm

        p_calls[0] += 1
        return 0

    f_calls = [0]
    p_calls = [0]

    xmin = fmin_lbfgs(f, [[10., 100.], [44., 55.]], progress, args=[f_calls])
    assert f_calls[0] > 0
    assert p_calls[0] > 0
    assert_array_almost_equal(xmin, [[0, 0], [0, 0]])


def test_class_interface():
    def f(x, g, *args):
        g[:] = 4 * x
        return x ** 4 + 1

    opt = LBFGS()
    opt.max_iterations = 3

    assert_array_equal(opt.minimize(f, 1e6), [0])

    opt.max_iterations = 1
    with pytest.raises(LBFGSError):
        opt.minimize(f, 1e7)


def test_input_validation():
    with pytest.raises(TypeError):
        fmin_lbfgs([], 1e4)
    with pytest.raises(TypeError):
        fmin_lbfgs(lambda x: x, 1e4, "ham")
    with pytest.raises(TypeError):
        fmin_lbfgs(lambda x: x, "spam")



import numpy as np

from sklearn.exceptions import NotFittedError
from sklearn.linear_model import Lasso, LinearRegression
from sklearn.linear_model._coordinate_descent import _alpha_grid
try:
    import pycasso
except ImportError:
    pycasso = None

from .base import AbstractUoILinearRegressor


class PycLasso():
    """Lasso using the pycasso solver. Solves for an entire regularization path
    at once.

    Parameters
    ----------
    alphas : nd-array
        The regularization path. Defaults to None for compatibility with UoI,
        but needs to be set prior to fitting.
    fit_intercept : bool
        Whether to calculate the intercept for this model. If set to ``False``,
        no intercept will be used in calculations.
    max_iter : int
        Maximum number of iterations for pycasso solver.

    Attributes
    ----------
    coef_ : ndarray, shape (n_features,) or (n_targets, n_features)
        Estimated coefficients for the linear regression problem.
    intercept_ : float
        Independent term in the linear model.
    """
    def __init__(self, alphas=None, fit_intercept=False, max_iter=1000):
        self.max_iter = max_iter
        self.fit_intercept = fit_intercept
        self.alphas = alphas

        # Flag to prevent us from predicting before fitting
        self.isfitted = False

    def set_params(self, **kwargs):
        """Sets the parameters of this estimator."""
        _valid_params = ['alphas', 'fit_intercept', 'max_iter']

        for key, value in kwargs.items():
            if key in _valid_params:
                setattr(self, key, value)
            else:
                raise ValueError('Invalid parameter %s' % key)

    def predict(self, X):
        """Predicts responses given a design matrix.

        Parameters
        ----------
        X : ndarray, (n_samples, n_features)
            The design matrix.

        Returns
        -------
        y : ndarray, shape (n_samples,)
            Predicted response vector.
        """
        if self.isfitted:
            return np.matmul(X, self.coef_.T) + self.intercept_
        else:
            raise NotFittedError('Estimator is not fit.')

    def fit(self, X, y):
        """Fit data according to the pycasso object.

        Parameters
        ----------
        X : ndarray, (n_samples, n_features)
            The design matrix.
        y : ndarray, shape (n_samples,)
            Response vector. Will be cast to X's dtype if necessary.
            Currently, this implementation does not handle multiple response
            variables.
        """
        if self.alphas is None:
            raise Exception('Set alphas before fitting.')

        self.solver = pycasso.Solver(X, y, family='gaussian',
                                     useintercept=self.fit_intercept,
                                     lambdas=self.alphas,
                                     penalty='l1',
                                     max_ite=self.max_iter)
        self.solver.train()
        # Coefs across the entire solution path
        self.coef_ = self.solver.result['beta']
        self.intercept_ = self.solver.result['intercept']
        self.isfitted = True
        return self


class UoI_Lasso(AbstractUoILinearRegressor, LinearRegression):
    r"""UoI\ :sub:`Lasso` solver.

    Parameters
    ----------
    n_boots_sel : int
        The number of data bootstraps/resamples to use in the selection module.
        Increasing this number will make selection more strict.
    n_boots_est : int
        The number of data bootstraps/resamples to use in the estimation
        module. Increasing this number will relax selection and decrease
        variance.
    n_lambdas : int
        The number of regularization values to use for selection.
    selection_frac : float
        The fraction of the dataset to use for training in each resampled
        bootstrap, during the selection module. Small values of this parameter
        imply larger "perturbations" to the dataset.
    estimation_frac : float
        The fraction of the dataset to use for training in each resampled
        bootstrap, during the estimation module. The remaining data is used
        to obtain validation scores. Small values of this parameters imply
        larger "perturbations" to the dataset.
    stability_selection : int, float, or array-like
        If int, treated as the number of bootstraps that a feature must
        appear in order to guarantee placement in selection profile. If float,
        must be between 0 and 1, and is instead the proportion of
        bootstraps. If array-like, must consist of either ints or floats
        between 0 and 1. In this case, each entry in the array-like object
        will act as a separate threshold for placement in the selection
        profile.
    estimation_score : string, "r2" | "AIC" | "AICc" | "BIC"
        Objective used to choose the best estimates per bootstrap.
    estimation_target : string, "train" | "test"
        Decide whether to assess the estimation_score on the train
        or test data across each bootstrap. By deafult, a sensible
        choice is made based on the chosen estimation_score
    warm_start : bool
        When set to ``True``, reuse the solution of the previous call to fit as
        initialization, otherwise, just erase the previous solution
    eps : float
        Length of the lasso path. ``eps=1e-3`` means that
        ``lambda_min / lambda_max = 1e-3``
    copy_X : bool
        If ``True``, X will be copied; else, it may be overwritten.
    fit_intercept : bool
        Whether to calculate the intercept for this model. If set
        to False, no intercept will be used in calculations
        (e.g. data is expected to be already centered).
    standardize : bool
        If True, the regressors X will be standardized before regression by
        subtracting the mean and dividing by their standard deviations. This
        parameter is equivalent to ``normalize`` in ``scikit-learn`` models.
    max_iter : int
        Maximum number of iterations for iterative fitting methods.
    random_state : int, RandomState instance, or None
        The seed of the pseudo random number generator that selects a random
        feature to update.  If int, random_state is the seed used by the random
        number generator; If RandomState instance, random_state is the random
        number generator; If None, the random number generator is the
        RandomState instance used by ``np.random``.
    comm : MPI communicator
        If passed, the selection and estimation steps are parallelized.
    logger : Logger
        The logger to use for messages when ``verbose=True`` in ``fit``.
        If *None* is passed, a logger that writes to ``sys.stdout`` will be
        used.
    solver : string, 'cd' | 'pyc'
        If cd, will use the ``scikit-learn`` lasso implementation (via
        coordinate descent). If pyc, will use pyclasso, built off of the
        pycasso path-wise solver.


    Attributes
    ----------
    coef_ : nd-array, shape (n_features,) or (n_targets, n_features)
        Estimated coefficients for the linear regression problem.
    intercept_ : float
        Independent term in the linear model.
    supports_ : array, shape
        boolean array indicating whether a given regressor (column) is selected
        for estimation for a given regularization parameter value (row).
    """
    def __init__(self, n_boots_sel=24, n_boots_est=24, selection_frac=0.9,
                 estimation_frac=0.9, n_lambdas=48, stability_selection=1.,
                 estimation_score='r2', estimation_target=None, eps=1e-3,
                 warm_start=True, copy_X=True, fit_intercept=True,
                 standardize=True, max_iter=1000, random_state=None,
                 comm=None, logger=None,
                 solver='cd'):
        super(UoI_Lasso, self).__init__(
            n_boots_sel=n_boots_sel,
            n_boots_est=n_boots_est,
            selection_frac=selection_frac,
            estimation_frac=estimation_frac,
            estimation_target=estimation_target,
            stability_selection=stability_selection,
            copy_X=copy_X,
            fit_intercept=fit_intercept,
            standardize=standardize,
            random_state=random_state,
            comm=comm,
            estimation_score=estimation_score,
            max_iter=max_iter,
            logger=logger)
        self.n_lambdas = n_lambdas
        self.eps = eps
        self.solver = solver

        if solver == 'cd':
            self._selection_lm = Lasso(
                max_iter=max_iter,
                warm_start=warm_start,
                random_state=random_state,
                fit_intercept=fit_intercept)
        elif solver == 'pyc':
            if pycasso is None:
                raise ImportError('pycasso is not installed.')
            self._selection_lm = PycLasso(
                fit_intercept=fit_intercept,
                max_iter=max_iter)

        self._estimation_lm = LinearRegression(fit_intercept=fit_intercept)

    def get_reg_params(self, X, y):
        alphas = _alpha_grid(
            X=X, y=y,
            l1_ratio=1.0,
            fit_intercept=self.fit_intercept,
            eps=self.eps,
            n_alphas=self.n_lambdas)

        return [{'alpha': a} for a in alphas]

    def uoi_selection_sweep(self, X, y, reg_param_values):
        """Overwrite base class selection sweep to accommodate pycasso
        path-wise solution"""

        if self.solver == 'pyc':
            alphas = np.array([reg_param['alpha']
                               for reg_param in reg_param_values])
            self._selection_lm.set_params(alphas=alphas)
            self._selection_lm.fit(X, y)

            return self._selection_lm.coef_
        else:
            return super(UoI_Lasso, self).uoi_selection_sweep(X, y,
                                                              reg_param_values)



import abc as _abc

from sklearn.linear_model._base import BaseEstimator


class AbstractDecompositionModel(BaseEstimator, metaclass=_abc.ABCMeta):
    @_abc.abstractmethod
    def fit(X):
        """Placeholder for fit. Subclasses should implement this method.
        Fit the model with X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        pass

    @_abc.abstractmethod
    def transform(self, X):
        """Apply dimensionality reduction to X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Data matrix to be transformed.

        Returns
        -------
        X_new : array-like, shape (n_samples, n_components)
            The transformed data matrix.
        """
        pass

    @_abc.abstractmethod
    def fit_transform(self, X):
        """Transform the data X according to the fitted decomposition.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Data matrix to be decomposed.

        Returns
        -------
        X_new : array-like, shape (n_samples, n_components)
            Transformed data.
        """
        pass
